
#include <linux/leds.h>
#include <mach/rma1evb_fpga.h>

#define FPGA_BASE	0x18000000

static void __iomem *fpga_base;
static DEFINE_SPINLOCK(fpga_lock);

u16 fpga_read(u32 offset)
{
	return __raw_readw(fpga_base + offset);
}
EXPORT_SYMBOL(fpga_read);

void fpga_write(u32 offset, u16 val, u16 mask)
{
	u16 tmp;
	unsigned long flags;

	spin_lock_irqsave(&fpga_lock, flags);
	tmp = fpga_read(offset);
	tmp = (tmp & ~mask) | (val & mask);
	__raw_writew(tmp, fpga_base + offset);
	spin_unlock_irqrestore(&fpga_lock, flags);
}
EXPORT_SYMBOL(fpga_write);

static void fpga_irq_mask(unsigned int irq)
{
	u32 irqpin = irq - FPGA_IRQ_BASE;
	u16 mask = 1 << (irqpin & 0xf);
	u32 offset = (irqpin & 0x10) ? FPGA_IRQMR1 : FPGA_IRQMR0;

	fpga_write(offset, mask, mask);
}

static void fpga_irq_unmask(unsigned int irq)
{
	u32 irqpin = irq - FPGA_IRQ_BASE;
	u16  mask = 1 << (irqpin & 0xf);
	u32 offset = (irqpin & 0x10) ? FPGA_IRQMR1 : FPGA_IRQMR0;

	fpga_write(offset, 0, mask);
}

static void fpga_demux(unsigned int irq, struct irq_desc *desc)
{
	struct irq_chip *chip = get_irq_chip(irq);
	u32 status, fpga_irq;
	u16 req0, req1;

	/* primary controller ack'ing */
	chip->ack(irq);

	do {
		req0 = fpga_read(FPGA_IRQSR0) & ~(fpga_read(FPGA_IRQMR0));
		req1 = fpga_read(FPGA_IRQSR1) & ~(fpga_read(FPGA_IRQMR1));

		status = (req1 << 16) | req0;
		if (status == 0)
			break;

		fpga_irq = FPGA_IRQ_BASE;
		while (status) {
			if (status & 1UL)
				generic_handle_irq(fpga_irq);
			status >>= 1;
			fpga_irq++;
		}
	} while (1);

	/* primary controller unmasking */
	chip->unmask(irq);
}

static struct irq_chip fpga_chip = {
	.name     = "FPGA",
	.mask     = fpga_irq_mask,
	.unmask   = fpga_irq_unmask,
	.disable  = fpga_irq_mask,
};

void __init fpga_init(void)
{
	int i;

	fpga_base = ioremap_nocache(FPGA_BASE, SZ_64K);

	/* VideoDec */
	fpga_write(FPGA_VDECCR, 1, 1);
	mdelay(10);	/* wait >=10ms before unreset VideoDec */

	/* Unreset Ether(2) */
	fpga_write(FPGA_DEVRSTCR1, 0x0004, 0x0004);
	/* Unreset AC2A(14)/Camera(12,11)/TP0(10)/eMMC(8)/VideoDec(7) */
	fpga_write(FPGA_DEVRSTCR2, 0x5d80, 0x5d80);
	/* VideoDec */
	fpga_write(FPGA_VDECCR, 3, 3);

	/* IRQ Mask */
	fpga_write(FPGA_IRQMR0, 0xffff, 0xffff);
	fpga_write(FPGA_IRQMR1, 0xffff, 0xffff);

	for (i = FPGA_IRQ_BASE; i < FPGA_IRQ_BASE + 32; i++) {
		set_irq_chip(i, &fpga_chip);
		set_irq_handler(i, handle_level_irq);
		set_irq_flags(i, IRQF_VALID);
	}
	set_irq_type(irqpin2irq(10), IRQ_TYPE_LEVEL_LOW);
	set_irq_chained_handler(irqpin2irq(10), fpga_demux);
}

/* LEDs */
struct fpga_led {
	struct led_classdev	cdev;
	u16			mask;
};

static const struct {
	const char *name;
	const char *trigger;
} fpga_leds[] = {
	{ "led0", },
	{ "led1", },
	{ "led2", },
	{ "led3", },
	{ "led4", "mmc0"},
	{ "led5", "mmc1"},
	{ "led6", "mmc2"},
	{ "led7", "sleep"},
};

static void fpga_led_set(struct led_classdev *cdev, enum led_brightness value)
{
	struct fpga_led *led = container_of(cdev, struct fpga_led, cdev);

	fpga_write(FPGA_LEDCR, value ? led->mask : 0, led->mask);
}

static enum led_brightness fpga_led_get(struct led_classdev *cdev)
{
	struct fpga_led *led = container_of(cdev, struct fpga_led, cdev);

	return (fpga_read(FPGA_LEDCR) & led->mask) ? LED_FULL : LED_OFF;
}

static int __init fpga_leds_init(void)
{
	int i;
	struct fpga_led *led;

	fpga_write(FPGA_LEDCR, 0, 0xffff);
	for (i = 0; i < ARRAY_SIZE(fpga_leds); i++) {
		led = kzalloc(sizeof(*led), GFP_KERNEL);
		if (!led)
			break;

		led->cdev.name = fpga_leds[i].name;
		led->cdev.brightness_set = fpga_led_set;
		led->cdev.brightness_get = fpga_led_get;
		led->cdev.default_trigger = fpga_leds[i].trigger;
		led->mask = BIT(i);

		if (led_classdev_register(NULL, &led->cdev) < 0) {
			kfree(led);
			break;
		}
	}
	return 0;
}
fs_initcall(fpga_leds_init);
