/*
 *  drivers/input/keyboard/rma1evb_btn.c
 *
 * Copyright (C) 2011 Renesas Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335, USA.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/io.h>
#include <linux/slab.h>
#include <linux/delay.h>

#include <linux/input.h>
#include <linux/input/rma1evb_btn.h>

#include <mach/rma1evb_fpga.h>

/* -------------------------------------------- */
/* constant definition				*/
/* -------------------------------------------- */
/* Resource Information */
#define MIN_SCAN_INTERVAL	30		/* 30ms */

/* FPGA_PSWCR */
#define KEY_CLEAR_MASK		0xFF00
#define KEY_STATE_MASK		0x00FF

#define DBG_PRINT(FMT, ARGS...)		\
		if (debug)		\
			printk(KERN_INFO "%s(): " FMT, __func__, ##ARGS)
#define DBG_PRINT2(FMT, ARGS...)	\
		if (debug > 1)		\
			printk(KERN_INFO "%s(): " FMT, __func__, ##ARGS)

/* -------------------------------------------- */
/* structure definition				*/
/* -------------------------------------------- */
struct rma1evb_btn_priv {
	int irq;
	u32 scan_interval;
	u16 last_keys;	/* last key state */
	struct platform_device *pdev;
	struct input_dev *input;
	int *keycodes;
};


/* -------------------------------------------- */
/* parameters definition			*/
/* -------------------------------------------- */
/* global parameters */
static u32 debug;
module_param(debug, int, 0644);
MODULE_PARM_DESC(debug, "RMA1EVB button debug level");

static inline u16 rma1evb_btn_read(u32 reg)
{
	return fpga_read(reg);
}

static inline void rma1evb_btn_write(u32 reg, u16 val, u16 mask)
{
	DBG_PRINT("reg =%04x val=%04x mask=%04x\n", reg, val, mask);
	fpga_write(reg, val, mask);
}

/*
 * rma1evb_btn_scan_core
 *    scan button matrix, and judge valid pressing
 */
static unsigned long rma1evb_btn_scan_core(struct rma1evb_btn_priv *priv)
{
	int i, mode;
	u16 key_event, chg_event, key_state;
	struct input_dev *key_dev = priv->input;

	key_state = rma1evb_btn_read(FPGA_PSWCR);
	key_event = key_state & KEY_STATE_MASK;
	if (priv->last_keys != key_event)
		DBG_PRINT("key event =0x%04x time=%lx\n", key_event, jiffies);

	chg_event = priv->last_keys ^ key_event;
	for (i = 0; chg_event; i++) {
		if (chg_event & 0x00000001) {
			mode = (key_event >> i) & 0x00000001;
			if (priv->keycodes[i] != KEY_RESERVED)
				input_event(key_dev, EV_KEY,
					priv->keycodes[i], mode);

			DBG_PRINT("%s sw%2d (%3d) time=%lx\n",
					mode ? "Press  " : "Release", i+24,
					priv->keycodes[i], jiffies);
		}
		chg_event >>= 1;
	}
	priv->last_keys = key_event;

	return key_state;
}

/*
 * rma1evb_btn_interrupt
 *     button interrupt handler for all buttons
 */
static irqreturn_t
rma1evb_btn_interrupt(int irq, void *dev_id)
{
	struct rma1evb_btn_priv *priv = dev_id;
	u16 key_event;

	key_event = rma1evb_btn_scan_core(priv);
	DBG_PRINT("irq %d, key_event %04x, time=%lx\n",
					irq, key_event, jiffies);
	while (key_event & KEY_STATE_MASK) {
		/* check key status */
		msleep(priv->scan_interval);
		key_event = rma1evb_btn_scan_core(priv);
	}

	/* interrupt status clear */
	rma1evb_btn_write(FPGA_PSWCR, key_event, KEY_CLEAR_MASK);

	return IRQ_HANDLED;
}


/*
 * rma1evb_btn_probe
 */
static int __devinit rma1evb_btn_probe(struct platform_device *pdev)
{
	struct rma1evb_btn_info *info;
	struct rma1evb_btn_priv *priv;
	int error, i;

	if (!pdev->dev.platform_data) {
		dev_err(&pdev->dev, "no platform data defined\n");
		return -EINVAL;
	}
	/* key code */
	info = (struct rma1evb_btn_info *)(pdev->dev.platform_data);

	error = -ENXIO;
	priv = kzalloc(sizeof(priv), GFP_KERNEL);
	if (priv == NULL) {
		dev_err(&pdev->dev, "failed to allocate driver data\n");
		return -ENOMEM;
	}

	priv->irq = platform_get_irq(pdev, 0);
	if (priv->irq < 0) {
		dev_err(&pdev->dev, "failed to get irq\n");
		return error;
	}

	platform_set_drvdata(pdev, priv);
	priv->pdev = pdev;
	priv->keycodes = (int *)&info->keycodes;

	/* setup input device */
	priv->input = input_allocate_device();
	if (!priv->input) {
		dev_err(&pdev->dev, "failed to allocate input device\n");
		error = -ENOMEM;
		goto err1;
	}
	priv->input->name = pdev->name;

	__set_bit(EV_KEY, priv->input->evbit);

	for (i = 0; i < RMA1EVB_MAXKEYS; i++) {
		if ((KEY_MAX >= priv->keycodes[i]) && priv->keycodes[i]) {
			__set_bit(priv->keycodes[i], priv->input->keybit);
			DBG_PRINT("key-%2d: %3d\n", i+24, priv->keycodes[i]);
		}
	}

	error = input_register_device(priv->input);
	if (error < 0) {
		dev_err(&pdev->dev, "Unable to register rma1evb_btn\n");
		goto err2;
	}

	/* set scan interval */
	if (info->scan_interval < MIN_SCAN_INTERVAL)
		priv->scan_interval = MIN_SCAN_INTERVAL;
	else
		priv->scan_interval = info->scan_interval;

	/* request interrupt handler */
	error = request_threaded_irq(priv->irq, NULL, rma1evb_btn_interrupt,
				     IRQF_ONESHOT, priv->input->name, priv);
	if (error) {
		dev_err(&pdev->dev, "%s(): reuest_irq(%d) failed (%d)\n",
						__func__, priv->irq, error);
		goto err3;
	}

	return 0;

 err3:
	input_unregister_device(priv->input);
 err2:
	input_free_device(priv->input);
 err1:
	platform_set_drvdata(pdev, NULL);
	kfree(priv);

	return error;
}

/*
 * rma1evb_btn_remove
 */
static int __devexit rma1evb_btn_remove(struct platform_device *pdev)
{
	struct rma1evb_btn_priv *priv = platform_get_drvdata(pdev);

	free_irq(priv->irq, priv);
	input_unregister_device(priv->input);
	input_free_device(priv->input);
	platform_set_drvdata(pdev, NULL);
	kfree(priv);

	return 0;
}

static struct platform_driver rma1evb_btn_driver = {
	.probe		= rma1evb_btn_probe,
	.remove		= __devexit_p(rma1evb_btn_remove),
	.driver		= {
		.name	= "rma1evb_btn",
		.owner	= THIS_MODULE,
	},
};

static int __init rma1evb_btn_init(void)
{
	return platform_driver_register(&rma1evb_btn_driver);
}

static void __exit rma1evb_btn_exit(void)
{
	platform_driver_unregister(&rma1evb_btn_driver);
}

module_init(rma1evb_btn_init);
module_exit(rma1evb_btn_exit);

MODULE_LICENSE("GPL");
