/*
 * Sample Driver for CE150x
 *
 * Copyright (C) 2011, Renesas Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/videodev2.h>
#include <linux/gpio.h>
#include <linux/firmware.h>

#include <media/soc_camera.h>
#include <media/soc_mediabus.h>
#include <media/v4l2-subdev.h>
#include <media/v4l2-chip-ident.h>

/* CE150X supported geometry for preview */
#define CE150X_WIDTH		640
#define CE150X_HEIGHT		480


struct ce150x_datafmt {
	enum v4l2_mbus_pixelcode	code;
	enum v4l2_colorspace		colorspace;
};

struct ce150x {
	struct v4l2_subdev		subdev;
	const struct ce150x_datafmt	*fmt;
};

static const struct ce150x_datafmt ce150x_colour_fmts[] = {
	{V4L2_MBUS_FMT_YUYV8_2X8, V4L2_COLORSPACE_SRGB},
};

static struct ce150x *
to_ce150x(const struct i2c_client *client)
{
	return container_of(i2c_get_clientdata(client), struct ce150x, subdev);
}

/* Find a data format by a pixel code in an array */
static const struct ce150x_datafmt *
ce150x_find_datafmt(enum v4l2_mbus_pixelcode code)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(ce150x_colour_fmts); i++)
		if (ce150x_colour_fmts[i].code == code)
			return ce150x_colour_fmts + i;

	return NULL;
}

static int
ce150x_try_fmt(struct v4l2_subdev *sd,
	       struct v4l2_mbus_framefmt *mf)
{
	const struct ce150x_datafmt *fmt = ce150x_find_datafmt(mf->code);

	dev_dbg(sd->v4l2_dev->dev, "%s(%u)\n", __func__, mf->code);

	if (!fmt) {
		mf->code	= ce150x_colour_fmts[0].code;
		mf->colorspace	= ce150x_colour_fmts[0].colorspace;
	}

	if (mf->width <= 640 && mf->height <= 480) {
		mf->width	= 640;
		mf->height	= 480;
	} else {
		mf->width	= CE150X_WIDTH;
		mf->height	= CE150X_HEIGHT;
		/*return -EINVAL;*/
	}
	mf->field	= V4L2_FIELD_NONE;

	return 0;
}

static int
ce150x_init_camera(struct v4l2_subdev *sd)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	u8 buf[2];

	/* Start Preview */
	buf[0] = 0x6B;
	buf[1] = 0x01;
	if (i2c_master_send(client, buf, sizeof(buf)) != sizeof(buf)) {
		dev_err(sd->v4l2_dev->dev, "i2c write error\n");
		return -EIO;
	}

	return 0;
}

static int
ce150x_s_fmt(struct v4l2_subdev *sd,
	     struct v4l2_mbus_framefmt *mf)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct ce150x *priv = to_ce150x(client);
	int ret;

	dev_dbg(sd->v4l2_dev->dev, "%s(%u)\n", __func__, mf->code);

	/* MIPI CSI could have changed the format, double-check */
	if (!ce150x_find_datafmt(mf->code)) {
		dev_err(sd->v4l2_dev->dev, "%s -EINVAL\n", __func__);
		return -EINVAL;
	}

	ret = ce150x_try_fmt(sd, mf);
	if (ret)
		return ret;

	priv->fmt = ce150x_find_datafmt(mf->code);

/*
	switch (mf->width) {
	case 320:
		break;
	case 640:
		break;
	case 1280:
		break;
	case 1920:
		break;
	case 2592:
		break;
	}
*/
	ce150x_init_camera(sd);

	return 0;
}

static int
ce150x_g_fmt(struct v4l2_subdev *sd,
	     struct v4l2_mbus_framefmt *mf)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct ce150x *priv = to_ce150x(client);

	const struct ce150x_datafmt *fmt = priv->fmt;

	mf->code	= fmt->code;
	mf->colorspace	= fmt->colorspace;
	mf->width	= CE150X_WIDTH;
	mf->height	= CE150X_HEIGHT;
	mf->field	= V4L2_FIELD_NONE;

	return 0;
}

static int
ce150x_g_crop(struct v4l2_subdev *sd, struct v4l2_crop *a)
{
	struct v4l2_rect *rect = &a->c;

	a->type		= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	rect->top	= 0;
	rect->left	= 0;
	rect->width	= CE150X_WIDTH;
	rect->height	= CE150X_HEIGHT;

	return 0;
}

static int
ce150x_cropcap(struct v4l2_subdev *sd, struct v4l2_cropcap *a)
{
	if (a->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	a->bounds.left			= 0;
	a->bounds.top			= 0;
	a->bounds.width			= CE150X_WIDTH;
	a->bounds.height		= CE150X_HEIGHT;
	a->defrect			= a->bounds;
	a->type				= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	a->pixelaspect.numerator	= 1;
	a->pixelaspect.denominator	= 1;

	return 0;
}

static int
ce150x_enum_fmt(struct v4l2_subdev *sd, unsigned int index,
		enum v4l2_mbus_pixelcode *code)
{
	if ((unsigned int)index >= ARRAY_SIZE(ce150x_colour_fmts))
		return -EINVAL;

	*code = ce150x_colour_fmts[index].code;
	return 0;
}

static int
ce150x_s_stream(struct v4l2_subdev *sd, int enable)
{
	struct i2c_client *client = sd->priv;
	struct soc_camera_device *icd = client->dev.platform_data;
	struct soc_camera_link *icl = to_soc_camera_link(icd);
	int ret;

	if (!icl->power)
		return -1;
	ret = icl->power(icd->pdev, enable ? 2 : 1);

	return ret;
}

static int
ce150x_g_chip_ident(struct v4l2_subdev *sd,
		    struct v4l2_dbg_chip_ident *id)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);

	if (id->match.type != V4L2_CHIP_MATCH_I2C_ADDR)
		return -EINVAL;

	if (id->match.addr != client->addr)
		return -ENODEV;

	id->ident	= V4L2_IDENT_UNKNOWN/*V4L2_IDENT_CE150X*/;
	id->revision	= 0;

	return 0;
}

static struct v4l2_subdev_video_ops ce150x_subdev_video_ops = {
	.s_stream	= ce150x_s_stream,
	.s_mbus_fmt	= ce150x_s_fmt,
	.g_mbus_fmt	= ce150x_g_fmt,
	.try_mbus_fmt	= ce150x_try_fmt,
	.enum_mbus_fmt	= ce150x_enum_fmt,
	.g_crop		= ce150x_g_crop,
	.cropcap	= ce150x_cropcap,
};

static struct v4l2_subdev_core_ops ce150x_subdev_core_ops = {
	.g_chip_ident	= ce150x_g_chip_ident,
};

static struct v4l2_subdev_ops ce150x_subdev_ops = {
	.core	= &ce150x_subdev_core_ops,
	.video	= &ce150x_subdev_video_ops,
};

static unsigned long
ce150x_query_bus_param(struct soc_camera_device *icd)
{
	struct soc_camera_link *icl = to_soc_camera_link(icd);
	unsigned long flags = SOCAM_PCLK_SAMPLE_RISING | SOCAM_MASTER |
		SOCAM_VSYNC_ACTIVE_HIGH | SOCAM_HSYNC_ACTIVE_HIGH |
		SOCAM_DATA_ACTIVE_HIGH;

	flags |= SOCAM_DATAWIDTH_16;

	return soc_camera_apply_sensor_flags(icl, flags);
}

static int
ce150x_set_bus_param(struct soc_camera_device *icd,
		     unsigned long flags)
{
	return 0;
}

static struct soc_camera_ops ce150x_ops = {
	.query_bus_param	= ce150x_query_bus_param,
	.set_bus_param		= ce150x_set_bus_param,
};

static int ce150x_probe(struct i2c_client *client,
			const struct i2c_device_id *did)
{
	struct ce150x *priv;
	struct soc_camera_device *icd = client->dev.platform_data;
	struct i2c_adapter *adapter = to_i2c_adapter(client->dev.parent);
	struct soc_camera_link *icl;
	int ret = 0;

	if (!icd) {
		dev_err(&client->dev, "CE150X: missing soc-camera data!\n");
		return -EINVAL;
	}

	icl = to_soc_camera_link(icd);
	if (!icl) {
		dev_err(&client->dev, "CE150X: missing platform data!\n");
		return -EINVAL;
	}

	if (!i2c_check_functionality(adapter, I2C_FUNC_I2C)) {
		dev_warn(&adapter->dev,
			 "I2C-Adapter doesn't support I2C_FUNC_I2C\n");
		return -EIO;
	}

	priv = kzalloc(sizeof(struct ce150x), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	v4l2_i2c_subdev_init(&priv->subdev, client, &ce150x_subdev_ops);

	icd->ops	= &ce150x_ops;
	priv->fmt	= &ce150x_colour_fmts[0];

	/* Start Camera FW */
	{
		u8 buf[1];

		buf[0] = 0xF0;
		if (i2c_master_send(client, buf, sizeof(buf)) != sizeof(buf)) {
			dev_err(&client->dev, "CE150X: i2c write error\n");
			ret = -EIO;
			goto failed;
		}
	}

	if (icl->free_bus)
		icl->free_bus(icl);


	return ret;
failed:
	icd->ops = NULL;
	kfree(priv);
	return ret;
}

static int ce150x_remove(struct i2c_client *client)
{
	struct ce150x *priv = to_ce150x(client);
	struct soc_camera_device *icd = client->dev.platform_data;
	struct soc_camera_link *icl = to_soc_camera_link(icd);

	icd->ops = NULL;
	if (icl->free_bus)
		icl->free_bus(icl);
	kfree(priv);

	return 0;
}

static const struct i2c_device_id ce150x_id[] = {
	{ "ce150x", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, ce150x_id);

static struct i2c_driver ce150x_i2c_driver = {
	.driver = {
		.name = "ce150x",
	},
	.probe		= ce150x_probe,
	.remove		= ce150x_remove,
	.id_table	= ce150x_id,
};

static int __init ce150x_mod_init(void)
{
	return i2c_add_driver(&ce150x_i2c_driver);
}

static void __exit ce150x_mod_exit(void)
{
	i2c_del_driver(&ce150x_i2c_driver);
}

module_init(ce150x_mod_init);
module_exit(ce150x_mod_exit);

MODULE_DESCRIPTION("CE150x sample driver");
MODULE_AUTHOR("Renesas Electronics Corporation");
MODULE_LICENSE("GPL v2");
