/*
 * drivers/video/sh_mobile_lcdcfb.c
 *
 * Copyright (C) 2011 Renesas Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
/*
 * SuperH Mobile LCDC Framebuffer
 *
 * Copyright (c) 2008 Magnus Damm
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */


#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/fb.h>
#include <linux/clk.h>
#include <linux/pm_runtime.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <linux/vmalloc.h>
#include <linux/ioctl.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <video/sh_mobile_lcdc.h>
#include <asm/atomic.h>

#if 0
#include <media/sh_mobile_appmem.h>
#endif

#ifdef CONFIG_HAS_EARLYSUSPEND
#include <linux/earlysuspend.h>
#endif /* CONFIG_HAS_EARLYSUSPEND */

#include <rtapi/screen_display.h>
#include <rtapi/screen_common.h>

#define CHAN_NUM 2

#define LAYER_NUM 4

#define PALETTE_NR 16
#define SIDE_B_OFFSET 0x1000
#define MIRROR_OFFSET 0x2000

struct sh_mobile_lcdc_priv;
struct sh_mobile_lcdc_chan {
	struct sh_mobile_lcdc_priv *lcdc;
	struct sh_mobile_lcdc_chan_cfg cfg;
	u32 pseudo_palette[PALETTE_NR];
	struct fb_info *info;
	dma_addr_t dma_handle;
	unsigned long pan_offset;
};

struct sh_mobile_lcdc_priv {
	struct device *dev;
	struct sh_mobile_lcdc_chan ch[CHAN_NUM];
#ifdef CONFIG_HAS_EARLYSUSPEND
	struct early_suspend    early_suspend;
#endif /* CONFIG_HAS_EARLYSUSPEND */

};

struct sh_mobile_lcdc_ext_param {
	struct semaphore sem_lcd;
	int lcd_type;
	void *aInfo;
	unsigned short o_mode;
	unsigned int phy_addr;
	unsigned int vir_addr;
	unsigned short rect_x;
	unsigned short rect_y;
	unsigned short rect_width;
	unsigned short rect_height;
	unsigned short alpha;
	unsigned short key_clr;
	unsigned short v4l2_state;
};

struct sh_mobile_lcdc_ext_param lcd_ext_param[CHAN_NUM];

struct sh_mobile_lcdc_layer_param {
	int lay_open;
	int lay_buffernum;
	int lay_num;
	int lay_overlay;
	unsigned short lay_alpha;
	unsigned short lay_pixfmt;
	long lay_keycolor;
	unsigned int lay_data_x;
	unsigned int lay_data_y;
	unsigned int lay_pos_x;
	unsigned int lay_pos_y;
	unsigned int lay_display_x;
	unsigned int lay_display_y;
	void *lay_handle;
	struct workqueue_struct *lay_wq;
	struct work_struct lay_work;
	unsigned char *lay_address;
	unsigned int lay_datasize;
	wait_queue_head_t lay_wait;
	int wait_state;
	int wait_skip;
	unsigned int lay_app_id;
	unsigned int lay_offset;
#if 0
	struct appmem_handle *app_handle[2];
	int handle_num;
	struct semaphore sem_app_free;
#endif
	int output_off;
};

struct sh_mobile_lcdc_layer_param lcd_layer_param[LAYER_NUM];

struct semaphore   sh_mobile_sem_open;

static int multi_layer_first;

static u32 fb_debug = 0;
#define DBG_PRINT(FMT, ARGS...)         \
	if (fb_debug)              \
		printk(KERN_INFO "%s(): " FMT, __func__, ##ARGS)

module_param(fb_debug, int, 0644);
MODULE_PARM_DESC(fb_debug, "SH LCD debug level");

static int sh_mobile_lcdc_setcolreg(u_int regno,
				    u_int red, u_int green, u_int blue,
				    u_int transp, struct fb_info *info)
{
	/* No. of hw registers */
	if (regno >= 256)
		return 1;

	/* grayscale works only partially under directcolor */
	if (info->var.grayscale) {
		/* grayscale = 0.30*R + 0.59*G + 0.11*B */
		red = green = blue = (red * 77 + green * 151 + blue * 28) >> 8;
	}

#define CNVT_TOHW(val, width) ((((val)<<(width))+0x7FFF-(val))>>16)
	switch (info->fix.visual) {
	case FB_VISUAL_TRUECOLOR:	/* FALL THROUGH */
	case FB_VISUAL_PSEUDOCOLOR:
	{
		red = CNVT_TOHW(red, info->var.red.length);
		green = CNVT_TOHW(green, info->var.green.length);
		blue = CNVT_TOHW(blue, info->var.blue.length);
		transp = CNVT_TOHW(transp, info->var.transp.length);
		break;
	}
	case FB_VISUAL_DIRECTCOLOR:
	{
		red = CNVT_TOHW(red, 8);	/* expect 8 bit DAC */
		green = CNVT_TOHW(green, 8);
		blue = CNVT_TOHW(blue, 8);
		/* hey, there is bug in transp handling... */
		transp = CNVT_TOHW(transp, 8);
		break;
	}
	}
#undef CNVT_TOHW
	/* Truecolor has hardware independent palette */
	if (info->fix.visual == FB_VISUAL_TRUECOLOR) {
		u32 v;

		if (regno >= 16)
			return 1;

		v = (red << info->var.red.offset) |
		    (green << info->var.green.offset) |
		    (blue << info->var.blue.offset) |
		    (transp << info->var.transp.offset);
		switch (info->var.bits_per_pixel) {
		case 16:	/* FALL THROUGH */
		case 24:	/* FALL THROUGH */
		case 32:
			((u32 *) (info->pseudo_palette))[regno] = v;
			break;
		case 8:		/* FALL THROUGH */
		default:
			break;
		}
	}

	return 0;
}

static struct fb_fix_screeninfo sh_mobile_lcdc_fix  = {
	.id		= "SH Mobile LCDC",
	.type		= FB_TYPE_PACKED_PIXELS,
	.visual		= FB_VISUAL_TRUECOLOR,
	.accel		= FB_ACCEL_NONE,
	.xpanstep	= 0,
#ifdef CONFIG_FB_SH_MOBILE_DOUBLE_BUF
	.ypanstep	= 1,
#else
	.ypanstep       = 0,
#endif
	.ywrapstep	= 0,
};

void sh_mobile_callback(int result, unsigned long user_data)
{
	return;
}


static int display_initialize(int lcd_num)
{
	screen_disp_new disp_new;
	screen_disp_param disp_param;
	screen_disp_get_address disp_addr;
	int ret = 0;

	disp_new.notify_display_lcd_blend
		= (void *)sh_mobile_callback;
	lcd_ext_param[lcd_num].aInfo = screen_display_new(&disp_new);
	if (lcd_ext_param[lcd_num].aInfo == NULL)
		return -2;

	disp_param.handle = lcd_ext_param[lcd_num].aInfo;
	disp_param.output_mode = lcd_ext_param[lcd_num].o_mode;
	disp_param.key_color = lcd_ext_param[lcd_num].key_clr;
	disp_param.alpha = lcd_ext_param[lcd_num].alpha;
        disp_param.blend_mode = RT_DISPLAY_BLEND_VIO;

	ret = screen_display_set_parameters(&disp_param);
	if (ret != SMAP_LIB_DISPLAY_OK) {
		printk(KERN_ALERT "disp_set_parameters err!\n");
		return -1;
	}

	disp_addr.handle = lcd_ext_param[lcd_num].aInfo;
	disp_addr.output_mode = lcd_ext_param[lcd_num].o_mode;
	ret = screen_display_get_address(&disp_addr);
	if (ret != SMAP_LIB_DISPLAY_OK) {
		printk(KERN_ALERT "disp_get_address err!\n");
		return -1;
	}
	return 0;
}

int sh_mobile_lcdc_keyclr_set(unsigned short s_key_clr,
			      unsigned short output_mode)
{
	int i, ret;
	screen_disp_new disp_new;
	screen_disp_param disp_param;
	screen_disp_delete disp_delete;
	void *screen_handle;

	for (i = 0 ; i < CHAN_NUM ; i++) {
		if (output_mode == lcd_ext_param[i].o_mode)
			break;
	}
	if (i >= CHAN_NUM) {
		printk(KERN_ALERT "lcdc_key_clr_set param ERR\n");
		return -1;
	}
	if (down_interruptible(&lcd_ext_param[i].sem_lcd)) {
		printk(KERN_ALERT "down_interruptible failed\n");
		return -1;
	}

	lcd_ext_param[i].key_clr = s_key_clr;

	if (lcd_ext_param[i].aInfo != NULL) {
		disp_new.notify_display_lcd_blend
			= (void *)sh_mobile_callback;
		screen_handle = screen_display_new(&disp_new);
		disp_param.handle = screen_handle;
		disp_param.output_mode = lcd_ext_param[i].o_mode;
		disp_param.key_color = lcd_ext_param[i].key_clr;
		disp_param.alpha = lcd_ext_param[i].alpha;
		disp_param.blend_mode = RT_DISPLAY_BLEND_VIO;
		ret = screen_display_set_parameters(&disp_param);
		if (ret != SMAP_LIB_DISPLAY_OK) {
			disp_delete.handle = screen_handle;
			screen_display_delete(&disp_delete);
			up(&lcd_ext_param[i].sem_lcd);
			return -1;
		}
		disp_delete.handle = screen_handle;
		screen_display_delete(&disp_delete);
	}

	up(&lcd_ext_param[i].sem_lcd);

	return 0;

}
EXPORT_SYMBOL(sh_mobile_lcdc_keyclr_set);

int sh_mobile_lcdc_alpha_set(unsigned short s_alpha,
			      unsigned short output_mode)
{
	int i, ret;
	screen_disp_new disp_new;
	screen_disp_param disp_param;
	screen_disp_delete disp_delete;
	void *screen_handle;

	for (i = 0 ; i < CHAN_NUM ; i++) {
		if (output_mode == lcd_ext_param[i].o_mode)
			break;
	}
	if (i >= CHAN_NUM) {
		printk(KERN_ALERT "lcdc_key_clr_set param ERR\n");
		return -1;
	}
	if (down_interruptible(&lcd_ext_param[i].sem_lcd)) {
		printk(KERN_ALERT "down_interruptible failed\n");
		return -1;
	}

	lcd_ext_param[i].alpha = s_alpha;

	if (lcd_ext_param[i].aInfo != NULL) {
		disp_new.notify_display_lcd_blend
			= (void *)sh_mobile_callback;
		screen_handle = screen_display_new(&disp_new);
		disp_param.handle = screen_handle;
		disp_param.output_mode = lcd_ext_param[i].o_mode;
		disp_param.key_color = lcd_ext_param[i].key_clr;
		disp_param.alpha = lcd_ext_param[i].alpha;
		disp_param.blend_mode = RT_DISPLAY_BLEND_VIO;
		ret = screen_display_set_parameters(&disp_param);
		if (ret != SMAP_LIB_DISPLAY_OK) {
			disp_delete.handle = screen_handle;
			screen_display_delete(&disp_delete);
			up(&lcd_ext_param[i].sem_lcd);
			return -1;
		}
		disp_delete.handle = screen_handle;
		screen_display_delete(&disp_delete);
	}

	up(&lcd_ext_param[i].sem_lcd);

	return 0;

}
EXPORT_SYMBOL(sh_mobile_lcdc_alpha_set);


int sh_mobile_lcdc_refresh(unsigned short set_state,
			      unsigned short output_mode)
{

#if 0
	int i, ret;
	screen_disp_set_lcd_refresh disp_refresh;

	for (i = 0 ; i < CHAN_NUM ; i++) {
		if (output_mode == lcd_ext_param[i].o_mode)
			break;
	}
	if (i >= CHAN_NUM) {
		printk(KERN_ALERT "lcdc_key_clr_set param ERR\n");
		up(&lcd_ext_param[i].sem_lcd);
		return -1;
	}
	if (down_interruptible(&lcd_ext_param[i].sem_lcd)) {
		printk(KERN_ALERT "down_interruptible failed\n");
		up(&lcd_ext_param[i].sem_lcd);
		return -1;
	}

	lcd_ext_param[i].v4l2_state = set_state;
	if (lcd_ext_param[i].aInfo != NULL) {
		disp_refresh.handle = lcd_ext_param[i].aInfo;
		disp_refresh.output_mode = lcd_ext_param[i].o_mode;
		disp_refresh.refresh_mode = set_state;
		ret = screen_display_set_lcd_refresh(&disp_refresh);
		if (ret != SMAP_LIB_DISPLAY_OK) {
			up(&lcd_ext_param[i].sem_lcd);
			return -1;
		}
	}

	up(&lcd_ext_param[i].sem_lcd);
#endif
	return 0;

}
EXPORT_SYMBOL(sh_mobile_lcdc_refresh);

static int sh_mobile_fb_pan_display(struct fb_var_screeninfo *var,
				     struct fb_info *info)
{
	struct sh_mobile_lcdc_chan *ch = info->par;
	unsigned long new_pan_offset;

	int ret = 0;

/* onscreen buffer 2 */
	unsigned int i;
	unsigned short set_format;
	unsigned char  lcd_num;
	screen_disp_draw disp_draw;

	new_pan_offset = (var->yoffset * info->fix.line_length) +
		(var->xoffset * (info->var.bits_per_pixel / 8));

	for (i = 0 ; i < CHAN_NUM ; i++) {
		if (ch->cfg.chan == lcd_ext_param[i].lcd_type)
			break;
	}
	lcd_num = i;
	if (lcd_num >= CHAN_NUM)
		return -EINVAL;

	if (down_interruptible(&lcd_ext_param[lcd_num].sem_lcd)) {
		printk(KERN_ALERT "down_interruptible failed\n");
		return -ERESTARTSYS;
	}

	/* Set the source address for the next refresh */

	if (lcd_ext_param[lcd_num].aInfo == NULL) {
		ret = display_initialize(lcd_num);
		if (ret == -1) {
			up(&lcd_ext_param[lcd_num].sem_lcd);
			return -EIO;
		} else if (ret == -2)
			printk(KERN_ALERT "nothing MFI driver\n");
	}
	if (lcd_ext_param[lcd_num].aInfo != NULL) {

		if (var->bits_per_pixel == 16)
			set_format = RT_DISPLAY_FORMAT_RGB565;
		else
			set_format = RT_DISPLAY_FORMAT_ARGB8888;

		disp_draw.handle = lcd_ext_param[lcd_num].aInfo;
		disp_draw.output_mode = lcd_ext_param[lcd_num].o_mode;
		disp_draw.draw_rect.x = lcd_ext_param[lcd_num].rect_x;
		disp_draw.draw_rect.y = lcd_ext_param[lcd_num].rect_y;
		disp_draw.draw_rect.width =
			lcd_ext_param[lcd_num].rect_width;
		disp_draw.draw_rect.height =
			lcd_ext_param[lcd_num].rect_height;
		disp_draw.format = set_format;
		disp_draw.buffer_offset = new_pan_offset;
		ret = screen_display_draw(&disp_draw);
		if (ret != SMAP_LIB_DISPLAY_OK) {
			DBG_PRINT("display_draw= %d\n",ret);
			up(&lcd_ext_param[lcd_num].sem_lcd);
			return -EIO;
		}
	}

	ch->pan_offset = new_pan_offset;

	up(&lcd_ext_param[lcd_num].sem_lcd);

	return 0;
}

void sh_mobile_callback_blend(int result, unsigned long user_data)
{

	DBG_PRINT("sh_mobile_callback_blend layer= %d\n",(int)user_data);

	if (lcd_layer_param[user_data].output_off == 0) {
		lcd_layer_param[user_data].wait_state = 1;
		wake_up_interruptible(&lcd_layer_param[user_data].lay_wait);
	} else {
		lcd_layer_param[user_data].output_off = 0;
		lcd_layer_param[user_data].wait_skip = 1;
	}
	return;

}

static void blend_open(struct work_struct *work)
{

	screen_disp_new disp_new;
	struct sh_mobile_lcdc_layer_param *lay_param =
		container_of(work, struct sh_mobile_lcdc_layer_param,
			     lay_work);

	DBG_PRINT("blend_open layer = %d\n",lay_param->lay_num);
	disp_new.notify_display_lcd_blend = (void *)sh_mobile_callback_blend;
	lay_param->lay_handle = screen_display_new(&disp_new);

	return;

}

static void blend_start(struct work_struct *work)
{

	screen_disp_lcd_blend_draw disp_blend_draw;
	struct sh_mobile_lcdc_layer_param *lay_param =
		container_of(work, struct sh_mobile_lcdc_layer_param,
			     lay_work);
	int ret;

	DBG_PRINT("blend_start layer = %d\n",lay_param->lay_num);

#if 0
	if (lay_param->lay_address == NULL) {
		if (down_interruptible(&lay_param->sem_app_free))
			printk(KERN_ALERT "down_interruptible failed\n");
		if (lay_param->app_handle[lay_param->handle_num & 0x1]
		    != NULL) {
			sh_mobile_appmem_free(
				lay_param->app_handle[
					lay_param->handle_num & 0x1]);
			lay_param->app_handle[lay_param->handle_num & 0x1]
				= NULL;
		}
		up(&lay_param->sem_app_free);
		lay_param->app_handle[lay_param->handle_num & 0x1]
			= sh_mobile_appmem_share(lay_param->lay_app_id, NULL);
		disp_blend_draw.address =
			sh_mobile_appmem_getaddress(
				lay_param->app_handle[lay_param->handle_num
						      & 0x1],
				lay_param->lay_offset);
		disp_blend_draw.apmem_handle =
			sh_mobile_appmem_getmemoryhandle(
				lay_param->app_handle[
					lay_param->handle_num & 0x1]);
	} else {
		lay_param->app_handle[lay_param->handle_num & 0x1] = NULL;
#endif
		disp_blend_draw.address = lay_param->lay_address;
		disp_blend_draw.apmem_handle = NULL;
#if 0
	}

	lay_param->handle_num++;
#endif
	disp_blend_draw.handle = lay_param->lay_handle;
	disp_blend_draw.layer = lay_param->lay_num;
	disp_blend_draw.layer_flag = RT_DISPLAY_OUTPUT_ON;

	disp_blend_draw.size = lay_param->lay_datasize;
	disp_blend_draw.rect.x = lay_param->lay_pos_x;
	disp_blend_draw.rect.y = lay_param->lay_pos_y;
	disp_blend_draw.rect.width = lay_param->lay_display_x;
	disp_blend_draw.rect.height = lay_param->lay_display_y;

	switch (lay_param->lay_pixfmt) {
	case SH_FB_DISPLAY_RGB565:
		disp_blend_draw.stride = lay_param->lay_data_x * 2;
		disp_blend_draw.format = RT_DISPLAY_FORMAT_RGB565;
		break;
	case SH_FB_DISPLAY_RGBA5551:
		disp_blend_draw.stride = lay_param->lay_data_x * 2;
		disp_blend_draw.format = RT_DISPLAY_FORMAT_RGBA5551;
		break;
	case SH_FB_DISPLAY_RGB888:
		disp_blend_draw.stride = lay_param->lay_data_x * 3;
		disp_blend_draw.format = RT_DISPLAY_FORMAT_RGB888;
		break;
	case SH_FB_DISPLAY_ARGB8888:
		disp_blend_draw.stride = lay_param->lay_data_x * 4;
		disp_blend_draw.format = RT_DISPLAY_FORMAT_ARGB8888;
		break;
	default:
		break;
	}


	disp_blend_draw.key_color = lay_param->lay_keycolor;
	disp_blend_draw.alpha = lay_param->lay_alpha;
	disp_blend_draw.user_data = lay_param->lay_num - 1;

	lay_param->wait_state = 0;

	DBG_PRINT("disp_blend_draw.handle = %x\n",
		  (unsigned int)disp_blend_draw.handle);
	DBG_PRINT("disp_blend_draw.layer = %d\n", disp_blend_draw.layer);
	DBG_PRINT("disp_blend_draw.layer_flag = %d\n",
		  disp_blend_draw.layer_flag);
	DBG_PRINT("disp_blend_draw.address = %x\n",
		  (unsigned int)disp_blend_draw.address);
	DBG_PRINT("disp_blend_draw.size = %d\n",
		  (unsigned int)disp_blend_draw.size);
	DBG_PRINT("disp_blend_draw.apmem_handle = %x\n",
		  (unsigned int)disp_blend_draw.apmem_handle);
	DBG_PRINT("disp_blend_draw.rect.x = %d\n",disp_blend_draw.rect.x);
	DBG_PRINT("disp_blend_draw.rect.y = %d\n",disp_blend_draw.rect.y);
	DBG_PRINT("disp_blend_draw.rect.width = %d\n",
		  disp_blend_draw.rect.width);
	DBG_PRINT("disp_blend_draw.rect.height = %d\n",
		  disp_blend_draw.rect.height);
	DBG_PRINT("disp_blend_draw.stride = %d\n",disp_blend_draw.stride);
	DBG_PRINT("disp_blend_draw.format = %d\n",disp_blend_draw.format);
	DBG_PRINT("disp_blend_draw.key_color = %x\n",
		  (unsigned int)disp_blend_draw.key_color);
	DBG_PRINT("disp_blend_draw.alpha = %x\n",disp_blend_draw.alpha);
	DBG_PRINT("disp_blend_draw.user_data = %d\n",
		  (unsigned int)disp_blend_draw.user_data);

	ret = screen_display_lcd_blend_draw(&disp_blend_draw);
	if(ret != SMAP_LIB_DISPLAY_OK)
		DBG_PRINT("ERR blend_draw = %d\n",ret);

	DBG_PRINT("blend_start END\n");
	return;

}

static void blend_outputoff(struct work_struct *work)
{

	screen_disp_lcd_blend_draw disp_blend_draw;
	int ret;
	struct sh_mobile_lcdc_layer_param *lay_param =
		container_of(work, struct sh_mobile_lcdc_layer_param,
			     lay_work);

	DBG_PRINT("blend_outputoff layer = %d\n",lay_param->lay_num);

	lay_param->output_off = 1;

	disp_blend_draw.handle = lay_param->lay_handle;
	disp_blend_draw.layer = lay_param->lay_num;
	disp_blend_draw.layer_flag = RT_DISPLAY_OUTPUT_OFF;

	DBG_PRINT("disp_blend_draw.handle = %x\n",
		  (unsigned int)disp_blend_draw.handle);
	DBG_PRINT("disp_blend_draw.layer = %d\n", disp_blend_draw.layer);
	DBG_PRINT("disp_blend_draw.layer_flag = %d\n",
		  disp_blend_draw.layer_flag);

	ret = screen_display_lcd_blend_draw(&disp_blend_draw);
	if(ret != SMAP_LIB_DISPLAY_OK)
		DBG_PRINT("ERR blend_draw = %d\n",ret);

	return;
}


static void blend_close(struct work_struct *work)
{

	screen_disp_delete disp_delete;
	struct sh_mobile_lcdc_layer_param *lay_param =
		container_of(work, struct sh_mobile_lcdc_layer_param,
			     lay_work);

	DBG_PRINT("blend_close layer = %d\n",lay_param->lay_num);

	if (lay_param->lay_handle != NULL) {
		disp_delete.handle = lay_param->lay_handle;
		screen_display_delete(&disp_delete);
		lay_param->lay_handle = NULL;
	}
#if 0
	if (down_interruptible(&lay_param->sem_app_free))
		printk(KERN_ALERT "down_interruptible failed\n");
	for(i= 0; i < 2; i++) {
		if (lay_param->app_handle[i] != NULL) {
			sh_mobile_appmem_free(lay_param->app_handle[i]);
			lay_param->app_handle[i] = NULL;
		}
	}
	up(&lay_param->sem_app_free);
#endif

	lay_param->lay_open = 0;

	return;
}

static int chk_supportpixfmt(int pixfmt)
{

	switch (pixfmt) {
	case SH_FB_DISPLAY_RGB565:
	case SH_FB_DISPLAY_RGB888:
	case SH_FB_DISPLAY_ARGB8888:
	case SH_FB_DISPLAY_RGBA5551:
		return 0;
	default:
		return -EINVAL;
	}

}

static int sh_mobile_ioctl(struct fb_info *info, unsigned int cmd,
		       unsigned long arg)
{
	int retval;

	switch (cmd) {
	case FBIO_WAITFORVSYNC:
		retval = 0;
		break;
	case FB_IOC_ISLAYEX:
		retval = 0;
		break;
	case FB_IOC_SUPPORTPIXFMT:
	{
		struct fb_lay_supportpixfmt sp_pixfmt;
		if (copy_from_user(&sp_pixfmt, (void *)arg,
				   sizeof(struct fb_lay_supportpixfmt))) {
			retval = -EFAULT;
			break;
		}
		retval = chk_supportpixfmt(sp_pixfmt.pixfmt);
		break;
	}
	case FB_IOCS_OPEN:
	{
		struct fb_lay_open open;
		int ret;
		if (copy_from_user(&open, (void *)arg,
				   sizeof(struct fb_lay_open))) {
			retval = -EFAULT;
			break;
		}
		if ((open.level >= LAYER_NUM) || (open.level < 0)) {
			retval = -EINVAL;
			break;
		}
		if ((open.overlay != SH_FB_BLEND_LCD) &&
		    (open.overlay != SH_FB_OVERLAY)) {
			retval = -EINVAL;
			break;
		}
		if (multi_layer_first == 0) {
			void *screen_handle;
			screen_disp_new disp_new;
			screen_disp_param disp_param;
			screen_disp_start_lcd disp_start_lcd;
			screen_disp_delete disp_delete;

			disp_new.notify_display_lcd_blend
				= (void *)sh_mobile_callback;
			screen_handle = screen_display_new(&disp_new);
			if (screen_handle == NULL) {
				printk(KERN_ALERT "nothing MFI driver\n");
				return -EIO;
			}
			disp_param.handle = screen_handle;
			disp_param.output_mode = RT_DISPLAY_LCD1;
			disp_param.key_color = lcd_ext_param[0].key_clr;
			disp_param.alpha = lcd_ext_param[0].alpha;
			disp_param.blend_mode = RT_DISPLAY_BLEND_LCD;

			ret = screen_display_set_parameters(&disp_param);
			if (ret != SMAP_LIB_DISPLAY_OK) {
				printk(KERN_ALERT
				       "disp_set_parameters err!\n");
				return -EIO;
			}
			disp_start_lcd.handle = screen_handle;
			disp_start_lcd.output_mode = RT_DISPLAY_LCD1;
			screen_display_start_lcd(&disp_start_lcd);

			disp_delete.handle = screen_handle;
			screen_display_delete(&disp_delete);

			multi_layer_first = 1;
		}
		if (down_interruptible(&sh_mobile_sem_open)) {
			printk(KERN_ALERT "down_interruptible failed\n");
			return -ERESTARTSYS;
		}
		if (lcd_layer_param[open.level].lay_open == 1) {
			retval = -EBUSY;
			up(&sh_mobile_sem_open);
			break;
		} else {
			lcd_layer_param[open.level].lay_open = 1;
			lcd_layer_param[open.level].lay_num = open.level + 1;
			lcd_layer_param[open.level].lay_overlay = open.overlay;
			if(open.overlay == SH_FB_BLEND_LCD) {
				lcd_layer_param[open.level].wait_skip = 1;
				lcd_layer_param[open.level].output_off = 0;
				lcd_layer_param[open.level].lay_buffernum
					= open.buffernum;
				lcd_layer_param[open.level].lay_alpha = 0;
				lcd_layer_param[open.level].lay_pixfmt
					= SH_FB_DISPLAY_ARGB8888;
				lcd_layer_param[open.level].lay_keycolor = 0;
				lcd_layer_param[open.level].lay_data_x = 0;
				lcd_layer_param[open.level].lay_data_y = 0;
				lcd_layer_param[open.level].lay_pos_x = 0;
				lcd_layer_param[open.level].lay_pos_y = 0;
				lcd_layer_param[open.level].lay_display_x = 0;
				lcd_layer_param[open.level].lay_display_y = 0;
				lcd_layer_param[open.level].lay_address = 0;
				lcd_layer_param[open.level].lay_datasize = 0;
#if 0
				lcd_layer_param[open.level].handle_num = 0;
				lcd_layer_param[open.level].app_handle[0]
					= NULL;
				lcd_layer_param[open.level].app_handle[1]
					= NULL;
#endif
				flush_workqueue(
					lcd_layer_param[open.level].lay_wq);
				PREPARE_WORK(
					&lcd_layer_param[open.level].lay_work,
					blend_open);
				queue_work(
					lcd_layer_param[open.level].lay_wq,
					&lcd_layer_param[open.level].lay_work);
				retval =0;
			} else {
				void *screen_handle;
				screen_disp_new disp_new;
				screen_disp_lcd_blend_draw disp_blend_draw;
				screen_disp_delete disp_delete;

				DBG_PRINT("overlay setting layer= %d\n",
					  lcd_layer_param[open.level].lay_num);

				disp_new.notify_display_lcd_blend
					= (void *)sh_mobile_callback;
				screen_handle = screen_display_new(&disp_new);
				disp_blend_draw.handle = screen_handle;
				disp_blend_draw.layer
					= lcd_layer_param[open.level].lay_num;
				disp_blend_draw.layer_flag
					= RT_DISPLAY_OUTPUT_MOVIE;
				ret = screen_display_lcd_blend_draw(
					&disp_blend_draw);
				if (ret == 0)
					retval = 0;
				else {
					DBG_PRINT("ERR lcd_blend_draw= %d\n",
						  ret);
					retval = -EINVAL;
				}
				disp_delete.handle = screen_handle;
				screen_display_delete(&disp_delete);
			}

				up(&sh_mobile_sem_open);
				break;
		}
	}
	case FB_IOCS_CLOSE:
	{
		struct fb_lay_close close;
		if (copy_from_user(&close, (void *)arg,
				   sizeof(struct fb_lay_close))) {
			retval = -EFAULT;
			break;
		}
		if ((close.level >= LAYER_NUM) || (close.level < 0) ||
			lcd_layer_param[close.level].lay_open == 0) {
			retval = -EINVAL;
			break;
		}

		if (lcd_layer_param[close.level].lay_overlay
		    == SH_FB_BLEND_LCD) {
			flush_workqueue(lcd_layer_param[close.level].lay_wq);
			PREPARE_WORK(&lcd_layer_param[close.level].lay_work,
				     blend_close);
			queue_work(lcd_layer_param[close.level].lay_wq,
				   &lcd_layer_param[close.level].lay_work);
		} else
			lcd_layer_param[close.level].lay_open = 0;

		retval =0;
		break;
	}
	case FB_IOCS_DISP:
	{
		struct fb_lay_disp disp;
		if (copy_from_user(&disp, (void *)arg,
				   sizeof(struct fb_lay_disp))) {
			retval = -EFAULT;
			break;
		}
		if ((disp.level >= LAYER_NUM) || (disp.level < 0) ||
		    lcd_layer_param[disp.level].lay_open == 0 ||
		    disp.on != SH_FB_DISP_OFF) {
			retval = -EINVAL;
			break;
		}
		flush_workqueue(lcd_layer_param[disp.level].lay_wq);
		PREPARE_WORK(&lcd_layer_param[disp.level].lay_work,
			     blend_outputoff);
		queue_work(lcd_layer_param[disp.level].lay_wq,
			   &lcd_layer_param[disp.level].lay_work);
		retval =0;
		break;
	}
	case FB_IOCS_ALPHA:
	{
		struct fb_lay_alpha s_alpha;
		if (copy_from_user(&s_alpha, (void *)arg,
				   sizeof(struct fb_lay_alpha))) {
			retval = -EFAULT;
			break;
		}
		if ((s_alpha.level >= LAYER_NUM) || (s_alpha.level < 0)) {
			retval = -EINVAL;
			break;
		}
		lcd_layer_param[s_alpha.level].lay_alpha = s_alpha.alpha;
		retval =0;
		break;
	}
	case FB_IOCG_ALPHA:
	{
		struct fb_lay_alpha g_alpha;
		if (copy_from_user(&g_alpha, (void *)arg,
				   sizeof(struct fb_lay_alpha))) {
			retval = -EFAULT;
			break;
		}
		if ((g_alpha.level >= LAYER_NUM) || (g_alpha.level < 0)) {
			retval = -EINVAL;
			break;
		}
		g_alpha.alpha = lcd_layer_param[g_alpha.level].lay_alpha;
		if (copy_to_user((void *)arg, &g_alpha,
				 sizeof(struct fb_lay_alpha))) {
			retval = -EFAULT;
			break;
		}
		retval =0;
		break;
	}
	case FB_IOCS_KEYCOLOR:
	{
		struct fb_lay_keycolor s_keycolor;
		if (copy_from_user(&s_keycolor, (void *)arg,
				   sizeof(struct fb_lay_keycolor))) {
			retval = -EFAULT;
			break;
		}
		if ((s_keycolor.level >= LAYER_NUM)
		    || (s_keycolor.level < 0)) {
			retval = -EINVAL;
			break;
		}
		lcd_layer_param[s_keycolor.level].lay_keycolor
			= s_keycolor.keycolor;
		retval =0;
		break;
	}
	case FB_IOCG_KEYCOLOR:
	{
		struct fb_lay_keycolor g_keycolor;
		if (copy_from_user(&g_keycolor, (void *)arg,
				   sizeof(struct fb_lay_keycolor))) {
			retval = -EFAULT;
			break;
		}
		if ((g_keycolor.level >= LAYER_NUM)
		    || (g_keycolor.level < 0)) {
			retval = -EINVAL;
			break;
		}
		g_keycolor.keycolor
			= lcd_layer_param[g_keycolor.level].lay_keycolor;
		if (copy_to_user((void *)arg, &g_keycolor,
				 sizeof(struct fb_lay_keycolor))) {
			retval = -EFAULT;
			break;
		}
		retval =0;
		break;
	}
	case FB_IOCS_LAYSIZE:
	{
		struct fb_lay_data_size s_data_size;
		if (copy_from_user(&s_data_size, (void *)arg,
				   sizeof(struct fb_lay_data_size))) {
			retval = -EFAULT;
			break;
		}
		if ((s_data_size.level >= LAYER_NUM)
		    || (s_data_size.level < 0)) {
			retval = -EINVAL;
			break;
		}

		lcd_layer_param[s_data_size.level].lay_data_x
			= s_data_size.x;
		lcd_layer_param[s_data_size.level].lay_data_y
			= s_data_size.y;
		retval =0;
		break;
	}
	case FB_IOCG_LAYSIZE:
	{
		struct fb_lay_data_size g_data_size;
		if (copy_from_user(&g_data_size, (void *)arg,
				   sizeof(struct fb_lay_data_size))) {
			retval = -EFAULT;
			break;
		}
		if ((g_data_size.level >= LAYER_NUM)
		    || (g_data_size.level < 0)) {
			retval = -EINVAL;
			break;
		}
		g_data_size.x
			= lcd_layer_param[g_data_size.level].lay_data_x;
		g_data_size.y
			= lcd_layer_param[g_data_size.level].lay_data_y;

		if (copy_to_user((void *)arg, &g_data_size,
				 sizeof(struct fb_lay_data_size))) {
			retval = -EFAULT;
			break;
		}
		retval =0;
		break;
	}
	case FB_IOCS_PIXFMT:
	{
		struct fb_lay_pixfmt s_pixfmt;
		if (copy_from_user(&s_pixfmt, (void *)arg,
				   sizeof(struct fb_lay_pixfmt))) {
			retval = -EFAULT;
			break;
		}
		if ((s_pixfmt.level >= LAYER_NUM)
		    || (s_pixfmt.level < 0)) {
			retval = -EINVAL;
			break;
		}
		if ((s_pixfmt.pixfmt > SH_FB_DISPLAY_RGBA5551)
		    || (s_pixfmt.pixfmt < SH_FB_DISPLAY_RGB565)) {
			retval = -EINVAL;
			break;
		}
		lcd_layer_param[s_pixfmt.level].lay_pixfmt
			= s_pixfmt.pixfmt;
		retval =0;
		break;
	}
	case FB_IOCG_PIXFMT:
	{
		struct fb_lay_pixfmt g_pixfmt;
		if (copy_from_user(&g_pixfmt, (void *)arg,
				   sizeof(struct fb_lay_pixfmt))) {
			retval = -EFAULT;
			break;
		}
		if ((g_pixfmt.level >= LAYER_NUM)
		    || (g_pixfmt.level < 0)) {
			retval = -EINVAL;
			break;
		}
		g_pixfmt.pixfmt
			= lcd_layer_param[g_pixfmt.level].lay_pixfmt;

		if (copy_to_user((void *)arg, &g_pixfmt,
				 sizeof(struct fb_lay_pixfmt))) {
			retval = -EFAULT;
			break;
		}
		retval =0;
		break;
	}
	case FB_IOCS_POS:
	{
		struct fb_lay_pos s_lay_pos;
		if (copy_from_user(&s_lay_pos, (void *)arg,
				   sizeof(struct fb_lay_pos))) {
			retval = -EFAULT;
			break;
		}
		if ((s_lay_pos.level >= LAYER_NUM)
		    || (s_lay_pos.level < 0)) {
			retval = -EINVAL;
			break;
		}
		if ((s_lay_pos.x >= SH_MLCD_WIDTH) || (s_lay_pos.x < 0) ||
		    (s_lay_pos.y >= SH_MLCD_HEIGHT) || (s_lay_pos.y < 0)) {
			retval = -EINVAL;
			break;
		}
		lcd_layer_param[s_lay_pos.level].lay_pos_x
			= s_lay_pos.x;
		lcd_layer_param[s_lay_pos.level].lay_pos_y
			= s_lay_pos.y;
		retval =0;
		break;
	}
	case FB_IOCG_POS:
	{
		struct fb_lay_pos g_lay_pos;
		if (copy_from_user(&g_lay_pos, (void *)arg,
				   sizeof(struct fb_lay_pos))) {
			retval = -EFAULT;
			break;
		}
		if ((g_lay_pos.level >= LAYER_NUM)
		    || (g_lay_pos.level < 0)) {
			retval = -EINVAL;
			break;
		}
		g_lay_pos.x
			= lcd_layer_param[g_lay_pos.level].lay_pos_x;
		g_lay_pos.y
			= lcd_layer_param[g_lay_pos.level].lay_pos_y;

		if (copy_to_user((void *)arg, &g_lay_pos,
				 sizeof(struct fb_lay_pos))) {
			retval = -EFAULT;
			break;
		}
		retval =0;
		break;
	}
	case FB_IOCS_SIZE:
	{
		struct fb_lay_display_size s_lay_disp_size;
		if (copy_from_user(&s_lay_disp_size, (void *)arg,
				   sizeof(struct fb_lay_display_size))) {
			retval = -EFAULT;
			break;
		}
		if ((s_lay_disp_size.level >= LAYER_NUM)
		    || (s_lay_disp_size.level < 0)) {
			retval = -EINVAL;
			break;
		}
		if ((s_lay_disp_size.x > SH_MLCD_WIDTH)
		    || (s_lay_disp_size.x < 1)
		    || (s_lay_disp_size.y > SH_MLCD_HEIGHT)
		    || (s_lay_disp_size.y < 1)) {
			retval = -EINVAL;
			break;
		}

		lcd_layer_param[s_lay_disp_size.level].lay_display_x
			= s_lay_disp_size.x;
		lcd_layer_param[s_lay_disp_size.level].lay_display_y
			= s_lay_disp_size.y;
		retval =0;
		break;
	}
	case FB_IOCG_SIZE:
	{
		struct fb_lay_display_size g_lay_disp_size;
		if (copy_from_user(&g_lay_disp_size, (void *)arg,
				   sizeof(struct fb_lay_display_size))) {
			retval = -EFAULT;
			break;
		}
		if ((g_lay_disp_size.level >= LAYER_NUM)
		    || (g_lay_disp_size.level < 0)) {
			retval = -EINVAL;
			break;
		}
		g_lay_disp_size.x
			= lcd_layer_param[g_lay_disp_size.level].lay_display_x;
		g_lay_disp_size.y
			= lcd_layer_param[g_lay_disp_size.level].lay_display_y;

		if (copy_to_user((void *)arg, &g_lay_disp_size,
				 sizeof(struct fb_lay_display_size))) {
			retval = -EFAULT;
			break;
		}
		retval =0;
		break;
	}
	case FB_IOCS_VIEWLAY:
	{
		struct fb_viewlay viewlay;
		if (copy_from_user(&viewlay, (void *)arg,
				   sizeof(struct fb_viewlay))) {
			retval = -EFAULT;
			break;
		}
		if ((viewlay.level >= LAYER_NUM)
		    || (viewlay.level < 0)) {
			retval = -EINVAL;
			break;
		}

		if ((viewlay.pos_x >= SH_MLCD_WIDTH) || (viewlay.pos_x < 0) ||
		    (viewlay.pos_y >= SH_MLCD_HEIGHT) || (viewlay.pos_x < 0)) {
			retval = -EINVAL;
			break;
		}

		if ((viewlay.display_x > SH_MLCD_WIDTH)
		    || (viewlay.display_x < 1)
		    || (viewlay.display_y > SH_MLCD_HEIGHT)
		    || (viewlay.display_y < 1)) {
			retval = -EINVAL;
			break;
		}

		lcd_layer_param[viewlay.level].lay_data_x
			= viewlay.data_x;
		lcd_layer_param[viewlay.level].lay_data_y
			= viewlay.data_y;
		lcd_layer_param[viewlay.level].lay_pos_x
			= viewlay.pos_x;
		lcd_layer_param[viewlay.level].lay_pos_y
			= viewlay.pos_y;
		lcd_layer_param[viewlay.level].lay_display_x
			= viewlay.display_x;
		lcd_layer_param[viewlay.level].lay_display_y
			= viewlay.display_y;
		retval =0;
		break;
	}
	case FB_IOCS_SWAPADDR:
	{
		struct fb_swapaddr s_swapaddr;
		if (copy_from_user(&s_swapaddr, (void *)arg,
				   sizeof(struct fb_swapaddr))) {
			DBG_PRINT("copy_from_user err\n");
			retval = -EFAULT;
			break;
		}
		if ((s_swapaddr.level >= LAYER_NUM) ||
		    (s_swapaddr.level < 0) ||
		    lcd_layer_param[s_swapaddr.level].lay_open == 0) {
			DBG_PRINT("level err\n");
			retval = -EINVAL;
			break;
		}
		if ((int)s_swapaddr.swapaddr & 0x1F) {
			DBG_PRINT("swapaddr Align err %x\n"
				  ,(int)s_swapaddr.swapaddr);
			retval = -EINVAL;
			break;
		}

		if (((lcd_layer_param[s_swapaddr.level].lay_pixfmt
		      == SH_FB_DISPLAY_RGB565) ||
		     (lcd_layer_param[s_swapaddr.level].lay_pixfmt
		      == SH_FB_DISPLAY_RGBA5551)) &&
		    (lcd_layer_param[s_swapaddr.level].lay_data_x & 0x1)) {
			DBG_PRINT("display size x err\n");
			retval = -EINVAL;
			break;
		}
		if ((lcd_layer_param[s_swapaddr.level].lay_pixfmt
		     == SH_FB_DISPLAY_RGB888) &&
		    (lcd_layer_param[s_swapaddr.level].lay_data_x & 0x3)) {
			DBG_PRINT("display size x err\n");
			retval = -EINVAL;
			break;
		}

		lcd_layer_param[s_swapaddr.level].lay_address
			= s_swapaddr.swapaddr;
		lcd_layer_param[s_swapaddr.level].lay_datasize
			= s_swapaddr.datasize;
#if 0
		lcd_layer_param[s_swapaddr.level].lay_app_id
			= s_swapaddr.app_id;
		lcd_layer_param[s_swapaddr.level].lay_offset
			= s_swapaddr.offset;
#endif
		flush_workqueue(lcd_layer_param[s_swapaddr.level].lay_wq);
		PREPARE_WORK(&lcd_layer_param[s_swapaddr.level].lay_work,
			     blend_start);
		queue_work(lcd_layer_param[s_swapaddr.level].lay_wq,
			   &lcd_layer_param[s_swapaddr.level].lay_work);
		retval =0;
		break;
	}
	case FB_IOC_WAITSWAP:
	{
		struct fb_waitswap waitswap;
		struct sh_mobile_lcdc_layer_param *layer_param;
		int ret;
		if (copy_from_user(&waitswap, (void *)arg,
				   sizeof(struct fb_waitswap))) {
			retval = -EFAULT;
			break;
		}
		if ((waitswap.level >= LAYER_NUM) || (waitswap.level < 0) ||
			lcd_layer_param[waitswap.level].lay_open == 0) {
			retval = -EINVAL;
			break;
		}
		layer_param = &lcd_layer_param[waitswap.level];

		if (layer_param->wait_skip == 1) {
			layer_param->wait_skip = 0;
			retval = 0;
			break;
		}
		ret = wait_event_interruptible_timeout(
			layer_param->lay_wait,
			layer_param->wait_state != 0,
			3 * HZ);
#if 0
		if (down_interruptible(&layer_param->sem_app_free)) {
			printk(KERN_ALERT "down_interruptible failed\n");
			return -ERESTARTSYS;
		}
		if (layer_param->app_handle[layer_param->handle_num & 0x1]
		    != NULL) {
			sh_mobile_appmem_free(
				layer_param->app_handle[
					layer_param->handle_num & 0x1]);
			layer_param->app_handle[layer_param->handle_num & 0x1]
				= NULL;
		}
		up(&layer_param->sem_app_free);
#endif
		if (ret == 0) {
			DBG_PRINT("TIMEDOUT err\n");
			retval = -ETIMEDOUT;
			break;
		} else if (ret < 0) {
			retval = ret;
			break;
		}
		retval = 0;
		break;
	}
	case FB_IOCG_LAYBUSY:
	{
		struct fb_lay_busy lay_busy;
		if (copy_from_user(&lay_busy, (void *)arg,
				   sizeof(struct fb_lay_busy))) {
			retval = -EFAULT;
			break;
		}
		if ((lay_busy.level >= LAYER_NUM)
		    || (lay_busy.level < 0)) {
			retval = -EINVAL;
			break;
		}

		if (lcd_layer_param[lay_busy.level].lay_open == 0)
			lay_busy.laybusy = SH_FB_LAY_NOBUSY;
		else
			lay_busy.laybusy = SH_FB_LAY_BUSY;

		if (copy_to_user((void *)arg, &lay_busy,
				 sizeof(struct fb_lay_busy))) {
			retval = -EFAULT;
			break;
		}
		retval =0;
		break;
	}

	default:
		retval = -ENOIOCTLCMD;
		break;
	}
	return retval;
}

static int sh_mobile_mmap(struct fb_info *info, struct vm_area_struct *vma)
{
	unsigned long start;
	unsigned long off;
	u32 len;

	if (vma->vm_pgoff > (~0UL >> PAGE_SHIFT))
		return -EINVAL;

	off = vma->vm_pgoff << PAGE_SHIFT;
	start = info->fix.smem_start;
	len = PAGE_ALIGN((start & ~PAGE_MASK) + info->fix.smem_len);

	if ((vma->vm_end - vma->vm_start + off) > len)
		return -EINVAL;

	off += start;
	vma->vm_pgoff = off >> PAGE_SHIFT;

	/* Accessing memory will be done non-cached. */
	vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);

	/* To stop the swapper from even considering these pages */
	vma->vm_flags |= (VM_IO | VM_RESERVED);

	if (remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
			    vma->vm_end - vma->vm_start, vma->vm_page_prot))
		return -EAGAIN;

	return 0;
}
static int sh_mobile_fb_check_var(struct fb_var_screeninfo *var,
				  struct fb_info *info)
{
	switch (var->bits_per_pixel) {
	case 16: /* RGB 565 */
		var->red.offset    = 11;
		var->red.length    = 5;
		var->green.offset  = 5;
		var->green.length  = 6;
		var->blue.offset   = 0;
		var->blue.length   = 5;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	case 32: /* ARGB 8888*/
		var->red.offset    = 16;
		var->red.length    = 8;
		var->green.offset  = 8;
		var->green.length  = 8;
		var->blue.offset   = 0;
		var->blue.length   = 8;
		var->transp.offset = 24;
		var->transp.length = 8;
		break;
	default:
		return -EINVAL;

	}
	return 0;

}

static struct fb_ops sh_mobile_lcdc_ops = {
	.owner          = THIS_MODULE,
	.fb_setcolreg	= sh_mobile_lcdc_setcolreg,
	.fb_check_var	= sh_mobile_fb_check_var,
	.fb_read        = fb_sys_read,
	.fb_write       = fb_sys_write,
	.fb_fillrect	= sys_fillrect,
	.fb_copyarea	= sys_copyarea,
	.fb_imageblit	= sys_imageblit,
	.fb_pan_display = sh_mobile_fb_pan_display,
	.fb_ioctl       = sh_mobile_ioctl,
	.fb_mmap	= sh_mobile_mmap,
};

static int sh_mobile_lcdc_set_bpp(struct fb_var_screeninfo *var, int bpp)
{
	switch (bpp) {
	case 16: /* RGB 565 */
		var->red.offset = 11;
		var->red.length = 5;
		var->green.offset = 5;
		var->green.length = 6;
		var->blue.offset = 0;
		var->blue.length = 5;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;

	case 32: /* ARGB 8888 */
		var->red.offset = 16;
		var->red.length = 8;
		var->green.offset = 8;
		var->green.length = 8;
		var->blue.offset = 0;
		var->blue.length = 8;
		var->transp.offset = 24;
		var->transp.length = 8;
		break;
	default:
		return -EINVAL;
	}
	var->bits_per_pixel = bpp;
	var->red.msb_right = 0;
	var->green.msb_right = 0;
	var->blue.msb_right = 0;
	var->transp.msb_right = 0;
	return 0;
}


static int sh_mobile_lcdc_suspend(struct device *dev)
{
	int lcd_num;
	screen_disp_new disp_new;
	screen_disp_stop_lcd disp_stop_lcd;
	screen_disp_delete disp_delete;
	void *suspend_handle;

	disp_new.notify_display_lcd_blend
		= (void *)sh_mobile_callback;
	suspend_handle = screen_display_new(&disp_new);

	for (lcd_num = 0; lcd_num < CHAN_NUM; lcd_num++) {
		if (lcd_ext_param[lcd_num].aInfo != NULL) {
			disp_stop_lcd.handle = suspend_handle;
			disp_stop_lcd.output_mode =
				lcd_ext_param[lcd_num].o_mode;
			screen_display_stop_lcd(&disp_stop_lcd);
		}
	}

	disp_delete.handle = suspend_handle;
	screen_display_delete(&disp_delete);

	return 0;
}

static int sh_mobile_lcdc_resume(struct device *dev)
{
	unsigned int lcd_num;
	screen_disp_new disp_new;
	screen_disp_start_lcd disp_start_lcd;
	screen_disp_delete disp_delete;
	void *suspend_handle;

	disp_new.notify_display_lcd_blend
		= (void *)sh_mobile_callback;
	suspend_handle = screen_display_new(&disp_new);

	for (lcd_num = 0; lcd_num < CHAN_NUM; lcd_num++) {
		if (lcd_ext_param[lcd_num].aInfo != NULL) {
			disp_start_lcd.handle = suspend_handle;
			disp_start_lcd.output_mode =
				lcd_ext_param[lcd_num].o_mode;
			screen_display_start_lcd(&disp_start_lcd);
		}
	}

	disp_delete.handle = suspend_handle;
	screen_display_delete(&disp_delete);

	return 0;
}

static const struct dev_pm_ops sh_mobile_lcdc_dev_pm_ops = {
	.suspend = sh_mobile_lcdc_suspend,
	.resume = sh_mobile_lcdc_resume,
};

#ifdef CONFIG_HAS_EARLYSUSPEND
static void sh_mobile_fb_early_suspend(struct early_suspend *h)
{
	struct sh_mobile_lcdc_priv *priv;

	priv = container_of(h, struct sh_mobile_lcdc_priv, early_suspend);

	sh_mobile_lcdc_suspend(priv->dev);

}

static void sh_mobile_fb_late_resume(struct early_suspend *h)
{
	struct sh_mobile_lcdc_priv *priv;

	priv = container_of(h, struct sh_mobile_lcdc_priv, early_suspend);

	sh_mobile_lcdc_resume(priv->dev);

}
#endif /* CONFIG_HAS_EARLYSUSPEND */

static int sh_mobile_lcdc_remove(struct platform_device *pdev);

static unsigned long RoundUpToMultiple(unsigned long x, unsigned long y)
{
	unsigned long div = x / y;
	unsigned long rem = x % y;

	return (div + ((rem == 0) ? 0 : 1)) * y;
}

#ifdef CONFIG_FB_SH_MOBILE_DOUBLE_BUF
static unsigned long GCD(unsigned long x, unsigned long y)
{
	while (y != 0) {
		unsigned long r = x % y;
		x = y;
		y = r;
	}
	return x;
}

static unsigned long LCM(unsigned long x, unsigned long y)
{
	unsigned long gcd = GCD(x, y);

	return (gcd == 0) ? 0 : ((x / gcd) * y);
}
#endif


static int __devinit sh_mobile_lcdc_probe(struct platform_device *pdev)
{
	struct fb_info *info;
	struct sh_mobile_lcdc_priv *priv;
	struct sh_mobile_lcdc_info *pdata;
	struct sh_mobile_lcdc_chan_cfg *cfg;
	struct resource *res;
	int error = 0;
	int i, j;
#ifdef CONFIG_FB_SH_MOBILE_DOUBLE_BUF
	unsigned long ulLCM;
#endif
	void *temp = NULL;
	char lay_buf[20];

	printk(KERN_ALERT "sh_mobile_lcdc_probe\n");

	if (!pdev->dev.platform_data) {
		dev_err(&pdev->dev, "no platform data defined\n");
		error = -EINVAL;
		goto err0;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);

	priv = kzalloc(sizeof(*priv), GFP_KERNEL);
	if (!priv) {
		dev_err(&pdev->dev, "cannot allocate device data\n");
		error = -ENOMEM;
		goto err0;
	}

	priv->dev = &pdev->dev;
	platform_set_drvdata(pdev, priv);
	pdata = pdev->dev.platform_data;

	j = 0;
	for (i = 0; i < ARRAY_SIZE(pdata->ch); i++) {
		priv->ch[j].lcdc = priv;
		memcpy(&priv->ch[j].cfg, &pdata->ch[i], sizeof(pdata->ch[i]));

		priv->ch[j].pan_offset = 0;
		switch (pdata->ch[i].chan) {
		case LCDC_CHAN_MAINLCD:
			lcd_ext_param[i].o_mode = RT_DISPLAY_LCD1;
			lcd_ext_param[i].rect_x = SH_MLCD_RECTX;
			lcd_ext_param[i].rect_y = SH_MLCD_RECTY;
			lcd_ext_param[i].rect_width = SH_MLCD_WIDTH;
			lcd_ext_param[i].rect_height = SH_MLCD_HEIGHT;
			lcd_ext_param[i].alpha = 0xFF;
			lcd_ext_param[i].key_clr = SH_MLCD_TRCOLOR;
			lcd_ext_param[i].v4l2_state = RT_DISPLAY_REFRESH_ON;
			lcd_ext_param[i].phy_addr = SCREEN_DISPLAY_BUFF_ADDR;
			j++;
			break;
		case LCDC_CHAN_SUBLCD:
			lcd_ext_param[i].o_mode = RT_DISPLAY_LCD2;
			lcd_ext_param[i].rect_x = SH_SLCD_RECTX;
			lcd_ext_param[i].rect_y = SH_SLCD_RECTY;
			lcd_ext_param[i].rect_width = SH_SLCD_WIDTH;
			lcd_ext_param[i].rect_height = SH_SLCD_HEIGHT;
			lcd_ext_param[i].alpha = 0xFF;
			lcd_ext_param[i].key_clr = SH_SLCD_TRCOLOR;
			lcd_ext_param[i].v4l2_state = RT_DISPLAY_REFRESH_ON;
			/* SUBLCD undefined */
			lcd_ext_param[i].phy_addr = 0;
			j++;
			break;
		}
	}

	if (!j) {
		dev_err(&pdev->dev, "no channels defined\n");
		error = -EINVAL;
		goto err1;
	}

	for (i = 0; i < j; i++) {
		cfg = &priv->ch[i].cfg;

		priv->ch[i].info = framebuffer_alloc(0, &pdev->dev);
		if (!priv->ch[i].info) {
			dev_err(&pdev->dev, "unable to allocate fb_info\n");
			error = -ENOMEM;
			break;
		}

		info = priv->ch[i].info;
		info->fbops = &sh_mobile_lcdc_ops;
		info->var.xres = info->var.xres_virtual = cfg->lcd_cfg.xres;
		info->var.yres = cfg->lcd_cfg.yres;
		/* Default Y virtual resolution is 2x panel size */
		info->var.width = cfg->lcd_size_cfg.width;
		info->var.height = cfg->lcd_size_cfg.height;
		info->var.activate = FB_ACTIVATE_NOW;
		error = sh_mobile_lcdc_set_bpp(&info->var, cfg->bpp);
		if (error)
			break;
		info->fix = sh_mobile_lcdc_fix;
		info->fix.line_length = cfg->lcd_cfg.xres * (cfg->bpp / 8);

#ifdef CONFIG_FB_SH_MOBILE_DOUBLE_BUF
		/* 4kbyte align */
		ulLCM = LCM(info->fix.line_length, 0x1000);
		info->fix.smem_len = RoundUpToMultiple(
			info->fix.line_length*info->var.yres, ulLCM);
		info->fix.smem_len *= 2;

		info->var.yres_virtual = info->fix.smem_len
			/ info->fix.line_length;
#else
		info->fix.smem_len = RoundUpToMultiple(
			info->fix.line_length*info->var.yres, 0x1000);
		info->var.yres_virtual = info->fix.smem_len
			/ info->fix.line_length;
#endif

/* onscreen buffer 2 */
		temp = ioremap(lcd_ext_param[i].phy_addr,
			       info->fix.smem_len);
		if (NULL == temp) {
			error = -ENOMEM;
			break;
		} else {
			lcd_ext_param[i].vir_addr = (unsigned int)temp;
		}

		info->pseudo_palette = &priv->ch[i].pseudo_palette;
		info->flags = FBINFO_FLAG_DEFAULT;

		error = fb_alloc_cmap(&info->cmap, PALETTE_NR, 0);
		if (error < 0) {
			dev_err(&pdev->dev, "unable to allocate cmap\n");
			break;
		}
		fb_set_cmap(&info->cmap, info);

/* onscreen buffer 2 */
		info->fix.smem_start = lcd_ext_param[i].phy_addr;
		info->screen_base = (char __iomem *)lcd_ext_param[i].vir_addr;
		info->device = &pdev->dev;
		info->par = &priv->ch[i];
	}

	if (error)
		goto err1;

	for (i = 0; i < j; i++) {
		struct sh_mobile_lcdc_chan *ch = priv->ch + i;

		info = ch->info;

		error = register_framebuffer(info);
		if (error < 0)
			goto err1;

		dev_info(info->dev,
			 "registered %s/%s as %dx%d %dbpp.\n",
			 pdev->name,
			 (ch->cfg.chan == LCDC_CHAN_MAINLCD) ?
			 "mainlcd" : "sublcd",
			 (int) ch->cfg.lcd_cfg.xres,
			 (int) ch->cfg.lcd_cfg.yres,
			 ch->cfg.bpp);

		lcd_ext_param[i].aInfo = NULL;
		lcd_ext_param[i].lcd_type = ch->cfg.chan;
		sema_init(&lcd_ext_param[i].sem_lcd, 1);
	}

	for (i = 0; i < LAYER_NUM; i++) {
		sprintf(lay_buf, "sh_mobile_lcd_lay%d", i);
		lcd_layer_param[i].lay_wq
			= create_singlethread_workqueue(lay_buf);
		INIT_WORK(&lcd_layer_param[i].lay_work, blend_open);
		init_waitqueue_head(&lcd_layer_param[i].lay_wait);
#if 0
		sema_init(&lcd_layer_param[i].sem_app_free, 1);
#endif
	}
	sema_init(&sh_mobile_sem_open, 1);
	multi_layer_first = 0;

#ifdef CONFIG_HAS_EARLYSUSPEND
	priv->early_suspend.level = EARLY_SUSPEND_LEVEL_DISABLE_FB;
	priv->early_suspend.suspend = sh_mobile_fb_early_suspend;
	priv->early_suspend.resume = sh_mobile_fb_late_resume;
	register_early_suspend(&priv->early_suspend);
#endif /* CONFIG_HAS_EARLYSUSPEND */

	return 0;
err1:
	sh_mobile_lcdc_remove(pdev);
err0:
	return error;
}

static int sh_mobile_lcdc_remove(struct platform_device *pdev)
{
	struct sh_mobile_lcdc_priv *priv = platform_get_drvdata(pdev);
	struct fb_info *info;
	int i;


#ifdef CONFIG_HAS_EARLYSUSPEND
	unregister_early_suspend(&priv->early_suspend);
#endif /* CONFIG_HAS_EARLYSUSPEND */

	for (i = 0; i < ARRAY_SIZE(priv->ch); i++)
		if (priv->ch[i].info->dev)
			unregister_framebuffer(priv->ch[i].info);

	for (i = 0; i < ARRAY_SIZE(priv->ch); i++) {
		info = priv->ch[i].info;

		if (!info || !info->device)
			continue;

		if (lcd_ext_param[i].vir_addr != 0)
			iounmap((void __iomem *)lcd_ext_param[i].vir_addr);

		fb_dealloc_cmap(&info->cmap);
		framebuffer_release(info);
	}

	kfree(priv);
	return 0;
}

static struct platform_driver sh_mobile_lcdc_driver = {
	.driver		= {
		.name		= "sh_mobile_lcdc_fb",
		.owner		= THIS_MODULE,
#if 0
		.pm		= &sh_mobile_lcdc_dev_pm_ops,
#endif
	},
	.probe		= sh_mobile_lcdc_probe,
	.remove		= sh_mobile_lcdc_remove,
};

static int __init sh_mobile_lcdc_init(void)
{
	return platform_driver_register(&sh_mobile_lcdc_driver);
}

static void __exit sh_mobile_lcdc_exit(void)
{
	platform_driver_unregister(&sh_mobile_lcdc_driver);
}

module_init(sh_mobile_lcdc_init);
module_exit(sh_mobile_lcdc_exit);

MODULE_DESCRIPTION("SuperH Mobile LCDC Framebuffer driver");
MODULE_AUTHOR("Renesas Electronics");
MODULE_LICENSE("GPL v2");
