/*
 * ak4648.c  --  AK4648 ALSA Soc Audio driver
 *
 * Copyright (C) 2011 Renesas Electronics Corporation.
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/i2c.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <sound/initval.h>
#include <sound/tlv.h>

#include <mach/rma1evb_pwic.h>

#include "ak4648.h"

#ifdef CONFIG_SND_SOC_SH4_FMSI_SRC
#include "../sh/fmsi_src.h"
#endif



/* #define AK4648_DEBUG */

#ifdef AK4648_DEBUG
#define dbg_msg(FMT, ARGS...) \
	printk(KERN_INFO "[Info]%s : "FMT, __func__, ##ARGS)
#else
#define dbg_msg(FMT, ARGS...) do {} while (0)
#endif




struct snd_soc_codec_device soc_codec_dev_ak4648;

/* codec private data */
struct ak4648_priv {
	struct snd_soc_codec codec;
	u8 reg_cache[AK4648_CACHEREGNUM];
	u8 dac_act, adc_act;
	u8 rate;
};

static struct snd_soc_codec *ak4648_codec;

/*
 * ak4648 register cache
 */
static const u8 ak4648_reg[AK4648_CACHEREGNUM] = {
	0x00,	/* AK4648_POWER_MANAGEMENT1		(0x00)	*/
	0x00,	/* AK4648_POWER_MANAGEMENT2		(0x01)	*/
	0x01,	/* AK4648_SIGNAL_SELECT1		(0x02)	*/
	0x00,	/* AK4648_SIGNAL_SELECT2		(0x03)	*/
	0x02,	/* AK4648_MODE_CONTROL1			(0x04)	*/
	0x00,	/* AK4648_MODE_CONTROL2			(0x05)	*/
	0x00,	/* AK4648_TIMER_SELECT			(0x06)	*/
	0x00,	/* AK4648_ALC_MODE_CONTROL1		(0x07)	*/
	0xe1,	/* AK4648_ALC_MODE_CONTROL2		(0x08)	*/
	0xe1,	/* AK4648_LCH_INPUT_VOLUME_CONTROL	(0x09)	*/
	0x18,	/* AK4648_LCH_DIGITA_VOLUME_CONTROL	(0x0a)	*/
	0x00,	/* AK4648_ALC_MODE_CONTROL3		(0x0b)	*/
	0xe1,	/* AK4648_RCH_INPUT_VOLUME_CONTROL	(0x0c)	*/
	0x18,	/* AK4648_RCH_DIGITA_VOLUME_CONTROL	(0x0d)	*/
	0x11,	/* AK4648_MODE_CONTROL3			(0x0e)	*/
	0xb8,	/* AK4648_MODE_CONTROL4			(0x0f)	*/
	0x00,	/* AK4648_POWER_MANAGEMENT3		(0x10)	*/
	0x00,	/* AK4648_DIGITAL_FILTER_SELECT		(0x11)	*/
	0x00,	/* AK4648_FIL3_COEFFICIENT0		(0x12)	*/
	0x00,	/* AK4648_FIL3_COEFFICIENT1		(0x13)	*/
	0x00,	/* AK4648_FIL3_COEFFICIENT2		(0x14)	*/
	0x00,	/* AK4648_FIL3_COEFFICIENT3		(0x15)	*/
	0x00,	/* AK4648_EQ_COEFFICIENT0		(0x16)	*/
	0x00,	/* AK4648_EQ_COEFFICIENT1		(0x17)	*/
	0x00,	/* AK4648_EQ_COEFFICIENT2		(0x18)	*/
	0x00,	/* AK4648_EQ_COEFFICIENT3		(0x19)	*/
	0x00,	/* AK4648_EQ_COEFFICIENT4		(0x1a)	*/
	0x00,	/* AK4648_EQ_COEFFICIENT5		(0x1b)	*/
	0x00,	/* AK4648_FIL1_COEFFICIENT0		(0x1c)	*/
	0x00,	/* AK4648_FIL1_COEFFICIENT1		(0x1d)	*/
	0x00,	/* AK4648_FIL1_COEFFICIENT2		(0x1e)	*/
	0x00,	/* AK4648_FIL1_COEFFICIENT3		(0x1f)	*/
	0x00,	/* AK4648_POWER_MANAGEMENT4		(0x20)	*/
	0x00,	/* AK4648_MODE_CONTROL5			(0x21)	*/
	0x00,	/* AK4648_LINEOUT_MIXING_SELECT		(0x22)	*/
	0x00,	/* AK4648_HP_MIXING_SELECT		(0x23)	*/
	0x00,	/* AK4648_SPK_MIXING_SELECT		(0x24)	*/
	0x88,	/* AK4648_EQ_CONTROL_250HZ_100HZ	(0x25)	*/
	0x88,	/* AK4648_EQ_CONTROL_3500HZ_1KHZ	(0x26)	*/
	0x08,	/* AK4648_EQ_CONTRO_10KHZ		(0x27)	*/
};


/*
 * digital output volume control:
 * from -115.0 to +12.0 dB in 0.5 dB steps
 */
static const DECLARE_TLV_DB_SCALE(out_tlv, -11500, 50, 1);


/*
 * digital input volume control:
 * from -54 to 36 dB in 0.375dB steps
 */
static const DECLARE_TLV_DB_SCALE(mic_amp_tlv, -5400, 38, 1);

static const struct snd_kcontrol_new ak4648_snd_controls[] = {
	/* Common playback gain controls */
	SOC_SINGLE_TLV("DAC Playback Volume",
			AK4648_LCH_DIGITA_VOLUME_CONTROL,
			0, 0xFF, 1, out_tlv),

	/* Common capture gain controls */
	SOC_SINGLE_TLV("Mic Amp Capture Volume",
			AK4648_LCH_INPUT_VOLUME_CONTROL,
			0, 0xF1, 0, mic_amp_tlv),

	/* SW Mute */
	SOC_SINGLE("Soft Mute",
			AK4648_MODE_CONTROL3, 5, 1, 0),
};


/* Output Mixers */
static const struct snd_kcontrol_new ak4648_out_mixer_controls[] = {
	SOC_DAPM_SINGLE("DACL", AK4648_SIGNAL_SELECT1,
			4, 1, 0),
};

/* Input MUXs */
static const char *ak4648_lin_mux_texts[] = {"LIN1"};
static const struct soc_enum ak4648_lin_mux_enum =
	SOC_ENUM_DOUBLE(AK4648_POWER_MANAGEMENT3, 1, 6,
			ARRAY_SIZE(ak4648_lin_mux_texts),
			ak4648_lin_mux_texts);
static const struct snd_kcontrol_new ak4648_lin_mux_control =
	SOC_DAPM_ENUM("Route", ak4648_lin_mux_enum);

static const char *ak4648_rin_mux_texts[] = {"RIN1"};
static const struct soc_enum ak4648_rin_mux_enum =
	SOC_ENUM_DOUBLE(AK4648_POWER_MANAGEMENT3, 2, 7,
			ARRAY_SIZE(ak4648_rin_mux_texts),
			ak4648_rin_mux_texts);
static const struct snd_kcontrol_new ak4648_rin_mux_control =
	SOC_DAPM_ENUM("Route", ak4648_rin_mux_enum);


static const struct snd_soc_dapm_widget ak4648_dapm_widgets[] = {

	SND_SOC_DAPM_PGA("PMMIN",
			AK4648_POWER_MANAGEMENT1, 5, 0, NULL, 0),

	/* DAC */
	SND_SOC_DAPM_DAC("DAC", "Playback",
			AK4648_POWER_MANAGEMENT1, 2, 0),

	/* Outputs */
	SND_SOC_DAPM_OUTPUT("OUT"),

	/* Output Mixers */
	/* for LOUT/ROUT */
	SND_SOC_DAPM_MIXER("OUT Mixer", AK4648_POWER_MANAGEMENT1,
			3, 0,
			&ak4648_out_mixer_controls[0],
			ARRAY_SIZE(ak4648_out_mixer_controls)),

	/* Inputs */
	SND_SOC_DAPM_INPUT("LIN1"),
	SND_SOC_DAPM_INPUT("RIN1"),

	/* ADC */
	SND_SOC_DAPM_ADC("ADC", "Capture",
			AK4648_POWER_MANAGEMENT1, 1, 0),

	/* Input MUXs */
	SND_SOC_DAPM_MUX("LIN MUX", AK4648_POWER_MANAGEMENT1,
			1, 0,
			&ak4648_lin_mux_control),
	SND_SOC_DAPM_MUX("RIN MUX", AK4648_POWER_MANAGEMENT1,
			1, 0,
			&ak4648_rin_mux_control),
};


static const struct snd_soc_dapm_route audio_map[] = {

	/* Outputs */
	{"OUT", "NULL", "PMMIN"},

	/* Output Mixer */
	{"PMMIN", "NULL", "OUT Mixer"},
	{"OUT Mixer", "DACL", "DAC"},

	/* Inputs */
	{"LIN MUX", "LIN1", "LIN1"},
	{"RIN MUX", "RIN1", "RIN1"},

	/* Inputs MUX*/
	{"ADC", "NULL", "LIN MUX"},
	{"ADC", "NULL", "RIN MUX"},
};

/*
 * read ak4648 register cache
 */
static inline unsigned int ak4648_read_reg_cache(struct snd_soc_codec *codec,
	unsigned int reg)
{
	u16 *cache = codec->reg_cache;
	if (reg >= AK4648_CACHEREGNUM)
		return -1;
	return cache[reg];
}

/*
 * write ak4648 register cache
 */
static inline void ak4648_write_reg_cache(struct snd_soc_codec *codec,
	u16 reg, unsigned int value)
{
	u16 *cache = codec->reg_cache;
	if (reg >= AK4648_CACHEREGNUM)
		return;

	cache[reg] = value;
}

/*
 * write to the AK4648 register space
 */
static int ak4648_write(struct snd_soc_codec *codec, unsigned int reg,
	unsigned int value)
{
	u8 data[2];

	/* data is
	 *   D15..D8 AK4648 register offset
	 *   D7...D0 register data
	 */
	data[0] = reg & 0xff;
	data[1] = value & 0xff;

	if (codec->hw_write(codec->control_data, data, 2) == 2) {
		ak4648_write_reg_cache(codec, reg, value);
		return 0;
	} else
		return -EIO;
}

static int ak4648_add_widgets(struct snd_soc_codec *codec)
{
	snd_soc_dapm_new_controls(codec, ak4648_dapm_widgets,
				  ARRAY_SIZE(ak4648_dapm_widgets));

	snd_soc_dapm_add_routes(codec, audio_map, ARRAY_SIZE(audio_map));

	return 0;
}



static int ak4648_sync(struct snd_soc_codec *codec)
{
	u16 *cache = codec->reg_cache;
	int i, r = 0;

	for (i = 0; i < AK4648_CACHEREGNUM; i++)
		r |= ak4648_write(codec, i, cache[i]);

	return r;
};


static int ak4648_dai_hw_params(struct snd_pcm_substream *substream,
				struct snd_pcm_hw_params *params,
				struct snd_soc_dai *dai)
{
	struct snd_soc_codec *codec = dai->codec;
	u8 rate;

	dbg_msg("hw_params\n");


#ifdef CONFIG_SND_SOC_SH4_FMSI_SRC
	switch (SRC_INPUT_RATE) {
#else
	switch (params_rate(params)) {
#endif
	case 7350:
		rate = AK4648_FS2;
		break;
	case 8000:
		rate = 0;
		break;
	case 11025:
		rate = AK4648_FS2 | AK4648_FS0;
		break;
	case 12000:
		rate = AK4648_FS0;
		break;
	case 14700:
		rate = AK4648_FS2 | AK4648_FS1;
		break;
	case 16000:
		rate = AK4648_FS1;
		break;
	case 22050:
		rate = AK4648_FS2 | AK4648_FS1 | AK4648_FS0;
		break;
	case 24000:
		rate = AK4648_FS1 | AK4648_FS0;
		break;
	case 29400:
		rate = AK4648_FS3 | AK4648_FS2 | AK4648_FS1;
		break;
	case 32000:
		rate = AK4648_FS3 | AK4648_FS1;
		break;
	case 44100:
		rate = AK4648_FS3 | AK4648_FS2 | AK4648_FS1 | AK4648_FS0;
		break;
	case 48000:
		rate = AK4648_FS3 | AK4648_FS1 | AK4648_FS0;
		break;
	default:
		return -EINVAL;
		break;
	}

	snd_soc_update_bits(codec, AK4648_MODE_CONTROL2, FS_MASK, rate);

	return 0;
}

static int ak4648_dai_set_sysclk(struct snd_soc_dai *codec_dai,
	int clk_id, unsigned int freq, int dir)
{
	struct snd_soc_codec *codec = codec_dai->codec;
	u8 pll;

	switch (freq) {
	case 11289600:
		pll = AK4648_PLL2;
		break;
	case 12288000:
		pll = AK4648_PLL2 | AK4648_PLL0;
		break;
	case 12000000:
		pll = AK4648_PLL2 | AK4648_PLL1;
		break;
	case 24000000:
		pll = AK4648_PLL2 | AK4648_PLL1 | AK4648_PLL0;
		break;
	case 13500000:
		pll = AK4648_PLL3 | AK4648_PLL2;
		break;
	case 27000000:
		pll = AK4648_PLL3 | AK4648_PLL2 | AK4648_PLL0;
		break;
	default:
		return -EINVAL;
	}
	snd_soc_update_bits(codec, AK4648_MODE_CONTROL1, PLL_MASK, pll);

	return 0;
}

static int ak4648_dai_set_fmt(struct snd_soc_dai *dai, unsigned int fmt)
{
	struct snd_soc_codec *codec = dai->codec;
	u8 data;
	u8 bcko;
	u8 dif;
	u8 ps;
	u8 msbs;
	u8 bckp;

	data = 0;
	bcko = 0;
	ps = 0;
	msbs = 0;
	bckp = 0;
	/* set master/slave audio interface */
	switch (fmt & SND_SOC_DAIFMT_MASTER_MASK) {
	case SND_SOC_DAIFMT_CBM_CFM:
		dbg_msg("SND_SOC_DAIFMT_CBM_CFM\n");

		data |= AK4648_PMPLL;
		data |= AK4648_MS;
		data |= AK4648_MCKO;
		bcko = BCKO_64;
		ps = AK4648_PS0;
		break;
	case SND_SOC_DAIFMT_CBS_CFS:
		dbg_msg("SND_SOC_DAIFMT_CBS_CFS\n");

		data &= ~AK4648_PMPLL;
		data &= ~AK4648_MS;
		data &= ~AK4648_MCKO;
		ps = 0;
		break;
	default:
		return -EINVAL;
	}

	switch (fmt & SND_SOC_DAIFMT_FORMAT_MASK) {
	case SND_SOC_DAIFMT_DSP_A: /* L data MSB after FRM LRC  */
		dbg_msg("SND_SOC_DAIFMT_DSP\n");

		dif = 0x00;
		msbs = 0;
		bckp = 0;
		break;
	case SND_SOC_DAIFMT_LSB: /* Right Justified mode */
		dbg_msg("SND_SOC_DAIFMT_LSB\n");

		dif = AK4648_DIF0;
		break;
	case SND_SOC_DAIFMT_MSB: /* Left Justified mode */
		dbg_msg("SND_SOC_DAIFMT_MSB\n");

		dif = AK4648_DIF1;
		break;
	case SND_SOC_DAIFMT_I2S:
		dbg_msg("SND_SOC_DAIFMT_I2S\n");

		dif = AK4648_DIF1 | AK4648_DIF0;
		break;
	default:
		return -EINVAL;
	}

	snd_soc_update_bits(codec, AK4648_MODE_CONTROL1,
					BCKO_MASK | DIF_MASK,
					bcko | dif);

	snd_soc_update_bits(codec, AK4648_MODE_CONTROL2,
					PS_MASK | AK4648_MSBS | AK4648_BCKP,
					ps | msbs | bckp);

	snd_soc_update_bits(codec, AK4648_POWER_MANAGEMENT2,
					AK4648_MS | AK4648_MCKO,
					data);

	return 0;
}

static int ak4648_set_bias_level(struct snd_soc_codec *codec,
		enum snd_soc_bias_level level)
{
	switch (level) {
	case SND_SOC_BIAS_ON:
		dbg_msg("SND_SOC_BIAS_ON\n");


#ifdef CONFIG_MACH_RMA1EVB
		pwic_write(PWIC_PVDD2, 0x9d);
#endif

		snd_soc_update_bits(codec, AK4648_POWER_MANAGEMENT1,
					AK4648_PMVCM, AK4648_PMVCM);
		break;
	case SND_SOC_BIAS_PREPARE:
		dbg_msg("SND_SOC_BIAS_PREPARE\n");
		break;
	case SND_SOC_BIAS_STANDBY:
		dbg_msg("SND_SOC_BIAS_STANDBY\n");

		snd_soc_update_bits(codec, AK4648_POWER_MANAGEMENT1,
					AK4648_PMVCM, 0);

#ifdef CONFIG_MACH_RMA1EVB
		pwic_write(PWIC_PVDD2, 0x00);
#endif

		break;
	case SND_SOC_BIAS_OFF:

		dbg_msg("ak4648_set_bias_level OFF %x\n", level);

		snd_soc_write(codec, AK4648_POWER_MANAGEMENT1, 0x00);
		break;
	}


	codec->bias_level = level;
	return 0;
}

static struct snd_soc_dai_ops ak4648_dai_ops = {
	.hw_params	= ak4648_dai_hw_params,
	.set_sysclk	= ak4648_dai_set_sysclk,
	.set_fmt	= ak4648_dai_set_fmt,
};

struct snd_soc_dai ak4648_dai = {
	.name = "AK4648",
	.playback = {
		.stream_name = "Playback",
		.channels_min = 2,
		.channels_max = 2,
		.rates = SNDRV_PCM_RATE_8000_48000 | SNDRV_PCM_RATE_KNOT,
		.formats = SNDRV_PCM_FMTBIT_S16_LE },
	.capture = {
		.stream_name = "Capture",
		.channels_min = 2,
		.channels_max = 2,
		.rates = SNDRV_PCM_RATE_8000_48000 | SNDRV_PCM_RATE_KNOT,
		.formats = SNDRV_PCM_FMTBIT_S16_LE },
	.ops = &ak4648_dai_ops,
	.symmetric_rates = 1,
};
EXPORT_SYMBOL_GPL(ak4648_dai);

static int ak4648_resume(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct snd_soc_codec *codec = socdev->card->codec;

	ak4648_sync(codec);
	return 0;
}

static int ak4648_probe(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct snd_soc_codec *codec;
	int ret = 0;

	if (ak4648_codec == NULL) {
		dev_err(&pdev->dev, "Codec device not registered\n");
		return -ENODEV;
	}

	socdev->card->codec = ak4648_codec;
	codec = ak4648_codec;

	/* register pcms */
	ret = snd_soc_new_pcms(socdev, SNDRV_DEFAULT_IDX1, SNDRV_DEFAULT_STR1);
	if (ret < 0) {
		dev_err(codec->dev, "failed to create pcms: %d\n", ret);
		goto pcm_err;
	}

	snd_soc_add_controls(codec, ak4648_snd_controls,
			     ARRAY_SIZE(ak4648_snd_controls));

	ak4648_add_widgets(codec);

	ak4648_set_bias_level(codec, SND_SOC_BIAS_STANDBY);


	snd_soc_write(codec, AK4648_LCH_INPUT_VOLUME_CONTROL, 0x91);
	snd_soc_write(codec, AK4648_RCH_INPUT_VOLUME_CONTROL, 0x91);

	snd_soc_update_bits(codec, AK4648_POWER_MANAGEMENT1,
					AK4648_PMADL, AK4648_PMADL);
	snd_soc_update_bits(codec, AK4648_POWER_MANAGEMENT3,
					AK4648_PMADR, AK4648_PMADR);

	dev_info(&pdev->dev, "AK4648 Audio Codec %s", AK4648_VERSION);
	return ret;

pcm_err:
	return ret;
}

/* power down chip */
static int ak4648_remove(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);

	snd_soc_free_pcms(socdev);
	snd_soc_dapm_free(socdev);

	return 0;
}

struct snd_soc_codec_device soc_codec_dev_ak4648 = {
	.probe =	ak4648_probe,
	.remove =	ak4648_remove,
	.resume =	ak4648_resume,
};
EXPORT_SYMBOL_GPL(soc_codec_dev_ak4648);

static int ak4648_register(struct ak4648_priv *ak4648,
		enum snd_soc_control_type control)
{
	struct snd_soc_codec *codec = &ak4648->codec;
	int ret = 0;

	ak4648->dac_act = 0;
	ak4648->adc_act = 0;


	if (ak4648_codec) {
		dev_err(codec->dev, "Another ak4648 is registered\n");
		ret = -EINVAL;
		goto err;
	}

	mutex_init(&codec->mutex);
	INIT_LIST_HEAD(&codec->dapm_widgets);
	INIT_LIST_HEAD(&codec->dapm_paths);

	snd_soc_codec_set_drvdata(codec, ak4648);
	codec->name		= "AK4648";
	codec->owner		= THIS_MODULE;
	codec->bias_level = SND_SOC_BIAS_STANDBY;
	codec->set_bias_level = ak4648_set_bias_level;
	codec->dai		= &ak4648_dai;
	codec->num_dai		= 1;
	codec->reg_cache_size = AK4648_CACHEREGNUM;
	codec->reg_cache = &ak4648->reg_cache;

	memcpy(codec->reg_cache, ak4648_reg, sizeof(ak4648_reg));

	ret = snd_soc_codec_set_cache_io(codec, 8, 8, control);
	if (ret < 0) {
		dev_err(codec->dev, "Failed to set cache I/O: %d\n", ret);
		goto err;
	}

	ak4648_dai.dev = codec->dev;
	ak4648_codec = codec;

	ret = snd_soc_register_codec(codec);
	if (ret) {
		dev_err(codec->dev, "Failed to register codec: %d\n", ret);
		goto err;
	}

	ret = snd_soc_register_dai(&ak4648_dai);
	if (ret) {
		dev_err(codec->dev, "Failed to register DAI: %d\n", ret);
		goto err_codec;
	}

	return 0;

err_codec:
	snd_soc_unregister_codec(codec);

err:
	kfree(ak4648);
	return ret;
}

static void ak4648_unregister(struct ak4648_priv *ak4648)
{
	ak4648_set_bias_level(&ak4648->codec, SND_SOC_BIAS_OFF);
	snd_soc_unregister_dai(&ak4648_dai);
	snd_soc_unregister_codec(&ak4648->codec);
	kfree(ak4648);
	ak4648_codec = NULL;
}

static int __devinit ak4648_i2c_probe(struct i2c_client *client,
			    const struct i2c_device_id *id)
{
	struct ak4648_priv *ak4648;
	struct snd_soc_codec *codec;

	ak4648 = kzalloc(sizeof(struct ak4648_priv), GFP_KERNEL);
	if (ak4648 == NULL)
		return -ENOMEM;

	codec = &ak4648->codec;
	codec->hw_write = (hw_write_t)i2c_master_send;

	i2c_set_clientdata(client, ak4648);
	codec->control_data = client;

	codec->dev = &client->dev;

	return ak4648_register(ak4648, SND_SOC_I2C);
}

static __devexit int ak4648_i2c_remove(struct i2c_client *client)
{
	struct ak4648_priv *ak4648 = i2c_get_clientdata(client);

	ak4648_unregister(ak4648);

	return 0;
}

static const struct i2c_device_id ak4648_i2c_id[] = {
	{ "ak4648", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, ak4648_i2c_id);

static struct i2c_driver ak4648_i2c_driver = {
	.driver = {
		.name = "AK4648 I2C Codec",
		.owner = THIS_MODULE,
	},
	.probe		= ak4648_i2c_probe,
	.remove		= __devexit_p(ak4648_i2c_remove),
	.id_table	= ak4648_i2c_id,
};

static int __init ak4648_modinit(void)
{
	return i2c_add_driver(&ak4648_i2c_driver);

}
module_init(ak4648_modinit);

static void __exit ak4648_exit(void)
{
	i2c_del_driver(&ak4648_i2c_driver);
}
module_exit(ak4648_exit);

MODULE_DESCRIPTION("ASoc AK4648 driver");
MODULE_AUTHOR("Renesas Electronics");
MODULE_LICENSE("GPL");
