/*
 * FIFO-incorporating Melody and Serial Interface (FMSI) support
 * 
 * Copyright (C) 2011 Renesas Electronics Corporation.
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "fmsi_spu_init_reg.h"

/* local variables */
static unsigned long init_state;

static unsigned long pbankc_init[DSP_NUM]; /* bankc initial value */
static unsigned long xbankc_init[DSP_NUM]; /* bankc initial value */
static unsigned long pbankc_mask;
static unsigned long xbankc_mask;

/* local macro function */
#define SPUSETR(ofs, v)    writel(v, spu2_base + ofs)
#define SPUGETR(ofs)       readl(spu2_base + ofs)

/*
 * This function must be called at system initialize,
 * at least before InitDSP()/QuitDSP().
 */
static irqreturn_t SPU_INT1(int irq, void *dev_id)
{
	dbg_msg_0("SPU_INT1\n");

	RSPF0_Interrupt(1);

	return IRQ_HANDLED;
}

static long SPU_Init(void __iomem *spu2_base_local)
{
	struct clk *spu_clk;
	spu_clk = clk_get(NULL, "spu_clk");
	if (IS_ERR(spu_clk)) {
		printk(KERN_ERR "no clock defined\n");
		return -1;
	}
	clk_set_rate(spu_clk,120000000);
	clk_enable(spu_clk);

	SPUSETR(SPUINIT_SPUSRST, 0x00000000);
	while (SPUGETR(SPUINIT_SPUSRST) & 0x00000001)
		;

	SPUSETR(SPUINIT_SPUSRST, 0x00000001);

	/* get initial value of BANK register */
	pbankc_mask = ~((1 << SPUINIT_TOTAL_PBANK_NUM) - 1);
	pbankc_init[0] = SPUGETR(SPUINIT_PBANKC0) & pbankc_mask;
	pbankc_init[1] = SPUGETR(SPUINIT_PBANKC1) & pbankc_mask;

	xbankc_mask = ~((1 << SPUINIT_TOTAL_XBANK_NUM) - 1);
	xbankc_init[0] = SPUGETR(SPUINIT_XBANKC0) & xbankc_mask;
	xbankc_init[1] = SPUGETR(SPUINIT_XBANKC1) & xbankc_mask;

	SPUSETR(SPUINIT_DSP0_DSPRST,		0x00000001);
	SPUSETR(SPUINIT_DSP1_DSPRST,		0x00000001);
	SPUSETR(SPUINIT_DSP0_DSPCORERST,	0x00000001);
	SPUSETR(SPUINIT_DSP1_DSPCORERST,	0x00000001);

	/* initialize PBANK register */
	SPUSETR(SPUINIT_PBANKC0, pbankc_init[0]);
	SPUSETR(SPUINIT_PBANKC1, pbankc_init[1]);

	/* initialize XBANK register */
	SPUSETR(SPUINIT_XBANKC0, xbankc_init[0]);
	SPUSETR(SPUINIT_XBANKC1, xbankc_init[1]);

	init_state = 1;

	return 0;
}


long SPU_InitDSP(void __iomem *spu2_base_local, unsigned long dsp_id,
		unsigned long pbank_num, unsigned long xbank_num)
{
	unsigned long p_used;
	unsigned long x_used;
	unsigned long p_setting;
	unsigned long x_setting;

	if (dsp_id >= DSP_NUM)
		return -1; /* parameter error */

	if (init_state != 1)
		return -2; /* state error */


	if (dsp_id == 0) {
		p_used = SPUGETR(SPUINIT_PBANKC1) & ~pbankc_mask;
		x_used = SPUGETR(SPUINIT_XBANKC1) & ~xbankc_mask;
		p_setting = ((1 << pbank_num) - 1) <<
				(SPUINIT_TOTAL_PBANK_NUM - pbank_num);
		x_setting = ((1 << xbank_num) - 1) <<
				(SPUINIT_TOTAL_XBANK_NUM - xbank_num);
	} else {
		p_used = SPUGETR(SPUINIT_PBANKC0) & ~pbankc_mask;
		x_used = SPUGETR(SPUINIT_XBANKC0) & ~xbankc_mask;
		p_setting = (1 << pbank_num) - 1;
		x_setting = (1 << xbank_num) - 1;
	}

	/* bank assignment */
	if (dsp_id == 0) {
		/* reset DSP */
		SPUSETR(SPUINIT_DSP0_DSPRST,		0x00000001);
		SPUSETR(SPUINIT_DSP0_DSPCORERST,	0x00000001);

		SPUSETR(SPUINIT_PBANKC0, pbankc_init[0] |  p_setting);
		SPUSETR(SPUINIT_XBANKC0, xbankc_init[0] |  x_setting);

	} else { /* dsp_id == 1 */
		/* reset DSP */
		SPUSETR(SPUINIT_DSP1_DSPRST,		0x00000001);
		SPUSETR(SPUINIT_DSP1_DSPCORERST,	0x00000001);

		SPUSETR(SPUINIT_PBANKC1, pbankc_init[1] |  p_setting);
		SPUSETR(SPUINIT_XBANKC1, xbankc_init[1] |  x_setting);

	}

	return 0;
}

long SPU_QuitDSP(void __iomem *spu2_base_local, unsigned long dsp_id)
{

	if (dsp_id >= DSP_NUM)
		return -1; /* parameter error */

	if (init_state != 1)
		return -2; /* state error */


	/* set bank register to initial */
	if (dsp_id == 0) {
		/* reset DSP */
		SPUSETR(SPUINIT_DSP0_DSPRST,		0x00000001);
		SPUSETR(SPUINIT_DSP0_DSPCORERST,	0x00000001);

		SPUSETR(SPUINIT_PBANKC0, pbankc_init[0]);
		SPUSETR(SPUINIT_XBANKC0, xbankc_init[0]);

	} else {
		/* reset DSP */
		SPUSETR(SPUINIT_DSP1_DSPRST,		0x00000001);
		SPUSETR(SPUINIT_DSP1_DSPCORERST,	0x00000001);

		SPUSETR(SPUINIT_PBANKC1, pbankc_init[1]);
		SPUSETR(SPUINIT_XBANKC1, xbankc_init[1]);

	}

	return 0;
}

long SPU_Quit(void __iomem *spu2_base_local)
{

	/* SPU - Software reset */
	SPUSETR(SPUINIT_SPUSRST, 0x00000000);
	while (SPUGETR(SPUINIT_SPUSRST) & 0x00000001)
		;

	init_state = 0;

	return 0;
}

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SuperH onchip FMSI audio driver");
MODULE_AUTHOR("Renesas Electronics");
