/*
 *  File Name       : RSPF0_spu_driver.c
 *  Function        : SPU OPEN SOUND I/F
 *  Release Version : Ver 1.00
 *  Release Date    : 2011/11/30
 *
 *  Copyright (C) 2011 Renesas Electronics Corporation
 *
 *  This program is free software;you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by Free
 *  Softwere Foundation; either version 2 of License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A
 *  PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program;
 *  If not, write to the Free Software Foundation, Inc., 51 Franklin St,
 *  Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
#include "RSPF0_spu_driver_reg.h"
#include "RSPF0_spu_driver.h"

/* definition */
#define COMM_STATE_BIT                  0x300000
#define COMM_LENGTH_BIT                 0x00001F
#define COMM_STATE_READY                0x000000
#define COMM_STATE_SEND                 0x100000
#define COMM_STATE_RECEIVE              0x200000

#define COMM_INDEX_API_ID               (0)
#define COMM_INDEX_API_PARAM_TOP        (1)

#define SPU_ADDR_SEND_PARAM(x)          (SPU_MEM_XTOP + ((x) << 2))
#define SPU_ADDR_RECEIVE_PARAM(x)       (SPU_MEM_XTOP + ((32+(x)) << 2))

#define SPUDRV_DSP_NUM                  2

/* local function */
static long download_program(unsigned long dsp_id, const unsigned long *data[3],
						const unsigned long size[3]);
static void write_word(unsigned long dsp_id, unsigned long addr,
							unsigned long data);
static void read_word(unsigned long dsp_id, unsigned long addr,
							unsigned long *data);
static void write_param(unsigned long dsp_id, unsigned long addr,
				unsigned long *data, unsigned long size);
static void read_param(unsigned long dsp_id, unsigned long addr,
				unsigned long *data, unsigned long size);

typedef struct {
	unsigned long boot_error;
	unsigned long init_flag;
	unsigned long address_mode;
	void (*api_receive_callback)(unsigned long, unsigned long *,
								unsigned long);
	void (*reg_read)(unsigned long *, unsigned long *, unsigned long size);
	void (*reg_write)(unsigned long *, unsigned long *, unsigned long size);
	long (*wait_for_event)(unsigned long);
	long (*set_event)(void);
	long (*enter_critical_section)(void);
	long (*leave_critical_section)(void);
} sSpuDrv;

static sSpuDrv spu_drv[SPUDRV_DSP_NUM];
static const unsigned long dsp_offset[2] = {0, SPU_DSP_ADDR_OFFSET};

/****************************************************************************
Function Name: (prefix)Open
Description:   initialize driver status and dsp
Parameter:
	unsigned long dsp_id       : the number of spu dsp to initialize
	void (*receive_callback)   : callback function. this function is
					called when dsp issues api to cpu.
	void (*reg_read)(...)      : this driver calls this callback to read
					from spu memory/registers.
	void (*reg_write)(...)     : this driver calls this callback to write
					to spu memory/registers.
	unsigned long download     : flag decides whether to download the
					program.
	unsigned long address_mode : 0 - 29bit mode, 1 - 32bit mode
	unsigned long *data[]      : top address of download data
	unsigned long size[]       : length of the download data
	unsigned long param[0]     : size of X-Memory
	unsigned long param[1]     : size of Y-Memory
	unsigned long param[2-8]   : extention parameter(fix to zero)
Returns:
	0   : success
	-1  : parameter error
	-2  : boot error
Notes:
	This function needs to be called after setting SPU bank registers.
****************************************************************************/
long SPUDRV_FUNC(Open)(
	unsigned long dsp_id,
	void (*receive_callback)(unsigned long, unsigned long *, unsigned long),
	SPUDRV_STRUCT(UserFunc) * func,
	unsigned long download,
	unsigned long address_mode,
	const unsigned long *data[3],
	const unsigned long size[3],
	unsigned long param[8])
{
	unsigned long i;
	unsigned long reg;

	/* parameter check */
	if (receive_callback == 0
		|| dsp_id >= SPUDRV_DSP_NUM
		|| func == 0
		|| func->read == 0
		|| func->write == 0
		|| func->set_event == 0
		|| func->wait_for_event == 0
		|| func->enter_critical_section == 0
		|| func->leave_critical_section == 0
		|| address_mode > 1
		|| data == 0
		|| size == 0
		|| param == 0
	 )
		return -1;

	spu_drv[dsp_id].boot_error = 0;
	spu_drv[dsp_id].init_flag = 0;
	spu_drv[dsp_id].address_mode = address_mode;
	spu_drv[dsp_id].api_receive_callback = receive_callback;
	spu_drv[dsp_id].reg_read = func->read;
	spu_drv[dsp_id].reg_write = func->write;
	spu_drv[dsp_id].set_event = func->set_event;
	spu_drv[dsp_id].wait_for_event = func->wait_for_event;
	spu_drv[dsp_id].enter_critical_section = func->enter_critical_section;
	spu_drv[dsp_id].leave_critical_section = func->leave_critical_section;

	/* reset dsp */
	write_word(dsp_id, SPU_REG_DSPCORERST, 0x00000001);
	write_word(dsp_id, SPU_REG_DSPRST,     0x00000001);
	write_word(dsp_id, SPU_REG_DSPRST,     0x00000000);

	if (download) {
		/* download P/X/Y Memory data */
		download_program(dsp_id, data, size);
	}

	/* enable spu interrupt */
	read_word(dsp_id, SPU_REG_IEMASKC, &reg);
	write_word(dsp_id, SPU_REG_IEMASKC, reg & ~0x00001000);
	read_word(dsp_id, SPU_REG_IMASKC, &reg);
	write_word(dsp_id, SPU_REG_IMASKC, reg & ~0x00001000);

	/* parameter range check */
	for (i = 0; i < 2; i++) {
		if (param[i] >= 65536)
			param[i] = 65535;
	}

	/* set parameters for dsp initizalization */
	write_param(dsp_id, SPU_ADDR_SEND_PARAM(0), (unsigned long *)param, 8);

	write_word(dsp_id, SPU_REG_COM0, 0x00500000 | dsp_id);

	/* start dsp */
	write_word(dsp_id, SPU_REG_DSPCORERST, 0x00000000);

	/* wait for dsp initizalization */

	if (spu_drv[dsp_id].wait_for_event(5000) < 0)
		return -3;


	/* check boot error */
	if (spu_drv[dsp_id].boot_error)
		return -2;
	else
		return 0;
}

/****************************************************************************
Function Name: (prefix)Close
Description:   terminate dsp and dma operation
Parameter:
	unsigned long dsp_id :
Returns:
	0   : success
	-1  : parameter error
	-2  : state error
	-3  : dma terminate error
Notes:
****************************************************************************/
long SPUDRV_FUNC(Close)(unsigned long dsp_id)
{
	unsigned long reg;
	unsigned long i;

	if (dsp_id >= SPUDRV_DSP_NUM)
		return -1;

	if (!spu_drv[dsp_id].init_flag)
		return -2;

	write_word(dsp_id, SPU_REG_DSPCORERST, 0x00000001);
	write_word(dsp_id, SPU_REG_CHCR0,      0x00000000);
	write_word(dsp_id, SPU_REG_CHCR1,      0x00000000);
	write_word(dsp_id, SPU_REG_CHCR2,      0x00000000);

	for (i = 0; i < 0x10000; i++) {
		read_word(dsp_id, SPU_REG_SPUSTS, &reg);
		if (!(reg & 0x00400000))
			break;
	}
	if (i == 0x10000) {
		/* DMA Terminate timeout occurred */
		return -3;
	}

	read_word(dsp_id, SPU_REG_IEMASKC, &reg);
	write_word(dsp_id, SPU_REG_IEMASKC, reg | 0x00001000);
	read_word(dsp_id, SPU_REG_IMASKC, &reg);
	write_word(dsp_id, SPU_REG_IMASKC, reg | 0x00001000);

	write_word(dsp_id, SPU_REG_DSPRST,     0x00000001);

	spu_drv[dsp_id].init_flag = 0;
	spu_drv[dsp_id].boot_error = 0;
	spu_drv[dsp_id].address_mode = 0;
	spu_drv[dsp_id].api_receive_callback = 0;
	spu_drv[dsp_id].reg_read = 0;
	spu_drv[dsp_id].reg_write = 0;
	spu_drv[dsp_id].set_event = 0;
	spu_drv[dsp_id].wait_for_event = 0;
	spu_drv[dsp_id].enter_critical_section = 0;
	spu_drv[dsp_id].leave_critical_section = 0;

	return 0;
}

/****************************************************************************
Function Name: (prefix)SendApi
Description:   send api based on SPU-framework to dsp
Parameter:
	int api    : API code
	int *data  : API paramter
	int size   : the length of the parameter[long word]
Returns:
	0   : success
	-1  : timeout occurred due to SPU busy
	-2  : user callback error
Notes:

****************************************************************************/
long SPUDRV_FUNC(SendApi)(unsigned long dsp_id, unsigned long api,
				unsigned long *data, unsigned long size) {
	unsigned long com;
	unsigned long wdata[17];
	unsigned long i;
	unsigned long k;

	for (k = 0; k < 0x8000; k++) {
		if (spu_drv[dsp_id].enter_critical_section() < 0)
			return -2; /* user callback error */
		read_word(dsp_id, SPU_REG_COM0, &com);

		/* if SPU is not busy, set API parameter */
		if ((com & COMM_STATE_BIT) != COMM_STATE_SEND) {
			com = COMM_STATE_SEND | (size+1);

			wdata[0] = api;
			for (i = 0; i < size; i++)
				wdata[i+1] = data[i];
			write_param(dsp_id, SPU_ADDR_SEND_PARAM(0), wdata,
									size+1);

			write_word(dsp_id, SPU_REG_COM0, com);

			write_word(dsp_id, SPU_REG_IESETD, 0x010000);

			if (spu_drv[dsp_id].leave_critical_section() < 0)
				return -2; /* user callback error */
			else
				return 0;
		}

		if (spu_drv[dsp_id].leave_critical_section() < 0)
			return -2; /* user callback error */
	}
	return -1; /* timeout */
}

/****************************************************************************
Function Name: (prefix)ISR1
Description:   interrupt service routine for the DSP1
Parameter:
	none
Returns:
	none
Notes:
Porting Notes:
	register this function at appropriate interrupt vector.
****************************************************************************/
void SPUDRV_FUNC(ISR)(unsigned long dsp_id)
{
	unsigned long cpu_com = 0, dsp_com = 0;
	unsigned long reg = 0;
	unsigned long ieventc;
	unsigned long length = 0;
	unsigned long param[17];

	read_word(dsp_id, SPU_REG_IEVENTC, &ieventc);

	if (spu_drv[dsp_id].init_flag == 0) {
		/* boot protocol */
		/* Boot process end notify from SPU */
		if (ieventc & 0x00001000) {
			write_word(dsp_id, SPU_REG_IEVENTC, ~0x00001000);

			read_word(dsp_id, SPU_REG_COM0, &cpu_com);

			if ((cpu_com & 0x00400000)) {
				if ((cpu_com & COMM_STATE_BIT) ==
				COMM_STATE_RECEIVE) {
					spu_drv[dsp_id].boot_error =
								cpu_com & 0xff;
					write_word(dsp_id, SPU_REG_COM0,
								0x00000000);
				} else
					spu_drv[dsp_id].boot_error = 0x80;
			} else
				spu_drv[dsp_id].boot_error = 0x81;
			spu_drv[dsp_id].init_flag = 1;
			if (spu_drv[dsp_id].set_event() != 0)
				spu_drv[dsp_id].boot_error = 0x82;
			return;
		}
	} else{
		/* api protocol */
		if (ieventc & 0x00001000) {
			write_word(dsp_id, SPU_REG_IEVENTC, ~0x00001000);

			read_word(dsp_id, SPU_REG_COM1, &dsp_com);

			/* api protocol(receive api from SPU) */
			if ((dsp_com & COMM_STATE_BIT) == COMM_STATE_SEND) {
				length = (dsp_com & COMM_LENGTH_BIT) - 1;

				read_param(dsp_id, SPU_ADDR_RECEIVE_PARAM(0),
							param, length + 1);

				write_word(dsp_id, SPU_REG_COM1,
				(reg & ~COMM_STATE_BIT) | COMM_STATE_RECEIVE);

				if (spu_drv[dsp_id].api_receive_callback)
					spu_drv[dsp_id].api_receive_callback(
					param[COMM_INDEX_API_ID],
					&param[COMM_INDEX_API_PARAM_TOP],
					length);
			}
		}
	}
}

/****************************************************************************
Function Name: download_program
Notes: this function is sub-function of Init()
****************************************************************************/
static long download_program(unsigned long dsp_id, const unsigned long *data[3],
						const unsigned long size[3]) {
	const unsigned long memtop_addr[3] = {
		SPU_MEM_PTOP,
		SPU_MEM_XTOP,
		SPU_MEM_YTOP
	};
	unsigned long i;
	unsigned long dest;

	for (i = 0; i < 3; i++) {
		dest = (dsp_offset[dsp_id] + memtop_addr[i]);
		spu_drv[dsp_id].reg_write((unsigned long *)dest,
					(unsigned long *)data[i], size[i]);
	}
	return 0;
}

static void write_word(unsigned long dsp_id, unsigned long addr,
							unsigned long data) {
	spu_drv[dsp_id].reg_write((unsigned long *)(dsp_offset[dsp_id] + addr),
	&data, 1);
}

static void read_word(unsigned long dsp_id, unsigned long addr,
							unsigned long *data) {
	spu_drv[dsp_id].reg_read((unsigned long *)(dsp_offset[dsp_id] + addr),
	data, 1);
}

static void read_param(unsigned long dsp_id, unsigned long addr,
					unsigned long *data, unsigned long size)
{
	unsigned long i;

	spu_drv[dsp_id].reg_read((unsigned long *)(dsp_offset[dsp_id] + addr),
	data, size);

	for (i = 0; i < size; i++)
		data[i] >>= 8;
}

static void write_param(unsigned long dsp_id, unsigned long addr,
					unsigned long *data, unsigned long size)
{
	unsigned long i;

	for (i = 0; i < size; i++)
		data[i] <<= 8;

	spu_drv[dsp_id].reg_write((unsigned long *)(dsp_offset[dsp_id] + addr),
	data, size);
}

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SPU OPEN SOUND I/F Ver 1.0");
MODULE_AUTHOR("Renesas Electronics");
